# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Replay buffers for Brax."""

import abc
import math
from typing import Generic, Optional, Sequence, Tuple, TypeVar

from brax.training.types import PRNGKey
import flax
import jax
from jax import flatten_util
from jax.experimental import pjit
import jax.numpy as jnp

State = TypeVar('State')
Sample = TypeVar('Sample')


class ReplayBuffer(abc.ABC, Generic[State, Sample]):
  """Contains replay buffer methods."""

  @abc.abstractmethod
  def init(self, key: PRNGKey) -> State:
    """Init the replay buffer."""

  def insert(self, buffer_state: State, samples: Sample) -> State:
    """Insert data into the replay buffer."""
    self.check_can_insert(buffer_state, samples, 1)
    return self.insert_internal(buffer_state, samples)

  def sample(self, buffer_state: State) -> Tuple[State, Sample]:
    """Sample a batch of data."""
    self.check_can_sample(buffer_state, 1)
    return self.sample_internal(buffer_state)

  def check_can_insert(self, buffer_state: State, samples: Sample, shards: int):
    """Checks whether insert can be performed. Do not JIT this method."""
    pass

  def check_can_sample(self, buffer_state: State, shards: int):
    """Checks whether sampling can be performed. Do not JIT this method."""
    pass

  @abc.abstractmethod
  def size(self, buffer_state: State) -> int:
    """Total amount of elements that are sampleable."""

  @abc.abstractmethod
  def insert_internal(self, buffer_state: State, samples: Sample) -> State:
    """Insert data into the replay buffer."""

  @abc.abstractmethod
  def sample_internal(self, buffer_state: State) -> Tuple[State, Sample]:
    """Sample a batch of data."""


@flax.struct.dataclass
class ReplayBufferState:
  """Contains data related to a replay buffer."""

  data: jnp.ndarray
  insert_position: jnp.ndarray
  sample_position: jnp.ndarray
  key: PRNGKey


class QueueBase(ReplayBuffer[ReplayBufferState, Sample], Generic[Sample]):
  """Base class for limited-size FIFO reply buffers.

  Implements an `insert()` method which behaves like a limited-size queue.
  I.e. it adds samples to the end of the queue and, if necessary, removes the
  oldest samples form the queue in order to keep the maximum size within the
  specified limit.

  Derived classes must implement the `sample()` method.
  """

  def __init__(
      self,
      max_replay_size: int,
      dummy_data_sample: Sample,
      sample_batch_size: int,
  ):
    self._flatten_fn = jax.vmap(lambda x: flatten_util.ravel_pytree(x)[0])

    dummy_flatten, self._unflatten_fn = flatten_util.ravel_pytree(
        dummy_data_sample
    )
    self._unflatten_fn = jax.vmap(self._unflatten_fn)
    data_size = len(dummy_flatten)

    self._data_shape = (max_replay_size, data_size)
    self._data_dtype = dummy_flatten.dtype
    self._sample_batch_size = sample_batch_size
    self._size = 0

  def init(self, key: PRNGKey) -> ReplayBufferState:
    return ReplayBufferState(
        data=jnp.zeros(self._data_shape, self._data_dtype),
        sample_position=jnp.zeros((), jnp.int32),
        insert_position=jnp.zeros((), jnp.int32),
        key=key,
    )

  def check_can_insert(self, buffer_state, samples, shards):
    """Checks whether insert operation can be performed."""
    assert isinstance(shards, int), 'This method should not be JITed.'
    insert_size = jax.tree_util.tree_flatten(samples)[0][0].shape[0] // shards
    if self._data_shape[0] < insert_size:
      raise ValueError(
          'Trying to insert a batch of samples larger than the maximum replay'
          f' size. num_samples: {insert_size}, max replay size'
          f' {self._data_shape[0]}'
      )
    self._size = min(self._data_shape[0], self._size + insert_size)

  def insert_internal(
      self, buffer_state: ReplayBufferState, samples: Sample
  ) -> ReplayBufferState:
    """Insert data in the replay buffer.

    Args:
      buffer_state: Buffer state
      samples: Sample to insert with a leading batch size.

    Returns:
      New buffer state.
    """
    if buffer_state.data.shape != self._data_shape:
      raise ValueError(
          f'buffer_state.data.shape ({buffer_state.data.shape}) '
          f"doesn't match the expected value ({self._data_shape})"
      )

    update = self._flatten_fn(samples)
    data = buffer_state.data

    # If needed, roll the buffer to make sure there's enough space to fit
    # `update` after the current position.
    position = buffer_state.insert_position
    roll = jnp.minimum(0, len(data) - position - len(update))
    data = jax.lax.cond(
        roll, lambda: jnp.roll(data, roll, axis=0), lambda: data
    )
    position = position + roll

    # Update the buffer and the control numbers.
    data = jax.lax.dynamic_update_slice_in_dim(data, update, position, axis=0)
    position = (position + len(update)) % (len(data) + 1)
    sample_position = jnp.maximum(0, buffer_state.sample_position + roll)

    return buffer_state.replace(
        data=data,
        insert_position=position,
        sample_position=sample_position,
    )

  def sample_internal(
      self, buffer_state: ReplayBufferState
  ) -> Tuple[ReplayBufferState, Sample]:
    raise NotImplementedError(f'{self.__class__}.sample() is not implemented.')

  def size(self, buffer_state: ReplayBufferState) -> int:
    return buffer_state.insert_position - buffer_state.sample_position  # pytype: disable=bad-return-type  # jax-ndarray


class Queue(QueueBase[Sample], Generic[Sample]):
  """Implements a limited-size queue replay buffer."""

  def __init__(
      self,
      max_replay_size: int,
      dummy_data_sample: Sample,
      sample_batch_size: int,
      cyclic: bool = False,
  ):
    """Initializes the queue.

    Args:
      max_replay_size: Maximum number of elements queue can have.
      dummy_data_sample: Example record to be stored in the queue, it is used to
        derive shapes.
      sample_batch_size: How many elements sampling from the queue should return
        in a batch.
      cyclic: Should sampling from the queue behave cyclicly, ie. once recently
        inserted element was sampled, sampling starts from the beginning of the
        buffer. For example, if the current queue content is [0, 1, 2] and
        `sample_batch_size` is 2, then consecutive calls to sample will give:
        [0, 1], [2, 0], [1, 2]...
    """
    super().__init__(max_replay_size, dummy_data_sample, sample_batch_size)
    self._cyclic = cyclic

  def check_can_sample(self, buffer_state, shards):
    """Checks whether sampling can be performed. Do not JIT this method."""
    assert isinstance(shards, int), 'This method should not be JITed.'
    if self._size < self._sample_batch_size:
      raise ValueError(
          f'Trying to sample {self._sample_batch_size * shards} elements, but'
          f' only {self._size * shards} available.'
      )
    if not self._cyclic:
      self._size -= self._sample_batch_size

  def sample_internal(
      self, buffer_state: ReplayBufferState
  ) -> Tuple[ReplayBufferState, Sample]:
    """Sample a batch of data.

    Args:
      buffer_state: Buffer state

    Returns:
      New buffer state and a batch with leading dimension 'sample_batch_size'.
    """
    if buffer_state.data.shape != self._data_shape:
      raise ValueError(
          f'Data shape expected by the replay buffer ({self._data_shape}) does '
          f'not match the shape of the buffer state ({buffer_state.data.shape})'
      )

    # Note that this may be out of bound, but the operations below would still
    # work fine as they take this number modulo the buffer size.
    idx = (jnp.arange(self._sample_batch_size) + buffer_state.sample_position) % buffer_state.insert_position

    flat_batch = jnp.take(buffer_state.data, idx, axis=0, mode='wrap')

    # Remove the sampled batch from the queue.
    sample_position = buffer_state.sample_position + self._sample_batch_size
    if self._cyclic:
      sample_position = sample_position % buffer_state.insert_position

    new_state = buffer_state.replace(sample_position=sample_position)
    return new_state, self._unflatten_fn(flat_batch)

  def size(self, buffer_state: ReplayBufferState) -> int:
    if self._cyclic:
      return buffer_state.insert_position  # pytype: disable=bad-return-type  # jax-ndarray
    else:
      return buffer_state.insert_position - buffer_state.sample_position  # pytype: disable=bad-return-type  # jax-ndarray


class UniformSamplingQueue(QueueBase[Sample], Generic[Sample]):
  """Implements an uniform sampling limited-size replay queue.

  * It behaves as a limited size queue (if buffer is full it removes the oldest
    elements when new one is inserted).
  * It supports batch insertion only (no single element)
  * It performs uniform random sampling with replacement of a batch of size
    `sample_batch_size`
  """

  def sample_internal(
      self, buffer_state: ReplayBufferState
  ) -> Tuple[ReplayBufferState, Sample]:
    if buffer_state.data.shape != self._data_shape:
      raise ValueError(
          f'Data shape expected by the replay buffer ({self._data_shape}) does '
          f'not match the shape of the buffer state ({buffer_state.data.shape})'
      )

    key, sample_key = jax.random.split(buffer_state.key)
    idx = jax.random.randint(
        sample_key,
        (self._sample_batch_size,),
        minval=buffer_state.sample_position,
        maxval=buffer_state.insert_position,
    )
    batch = jnp.take(buffer_state.data, idx, axis=0, mode='wrap')
    return buffer_state.replace(key=key), self._unflatten_fn(batch)


class PmapWrapper(ReplayBuffer[State, Sample]):
  """Wrapper to distribute the buffer on multiple devices.

  Each device stores a replay buffer 'buffer' such that no data moves from one
  device to another.
  The total capacity of this replay buffer is the number of devices multiplied
  by the size of the wrapped buffer.
  The sample size is also the number of devices multiplied by the size of the
  wrapped buffer.
  This should not be used inside a pmapped function:
  You should just use the regular replay buffer in that case.
  """

  def __init__(
      self,
      buffer: ReplayBuffer[State, Sample],
      local_device_count: Optional[int] = None,
  ):
    self._buffer = buffer
    self._num_devices = local_device_count or jax.local_device_count()

  def init(self, key: PRNGKey) -> State:
    key = jax.random.fold_in(key, jax.process_index())
    keys = jax.random.split(key, self._num_devices)
    return jax.pmap(self._buffer.init)(keys)

  # NB: In multi-hosts setups, every host is expected to give a different batch.
  def insert(self, buffer_state: State, samples: Sample) -> State:
    self._buffer.check_can_insert(buffer_state, samples, self._num_devices)
    samples = jax.tree_util.tree_map(
        lambda x: jnp.reshape(x, (-1, self._num_devices) + x.shape[1:]), samples
    )
    # This is to enforce we're gonna iterate on the start of the batch before
    # the end of the batch.
    samples = jax.tree_util.tree_map(lambda x: jnp.swapaxes(x, 0, 1), samples)
    return jax.pmap(self._buffer.insert_internal)(buffer_state, samples)

  # NB: In multi-hosts setups, every host will get a different batch.
  def sample(self, buffer_state: State) -> Tuple[State, Sample]:
    self._buffer.check_can_sample(buffer_state, self._num_devices)
    buffer_state, samples = jax.pmap(self._buffer.sample_internal)(buffer_state)
    samples = jax.tree_util.tree_map(lambda x: jnp.swapaxes(x, 0, 1), samples)
    samples = jax.tree_util.tree_map(
        lambda x: jnp.reshape(x, (-1,) + x.shape[2:]), samples
    )
    return buffer_state, samples

  def insert_internal(self, buffer_state: State, samples: Sample) -> State:
    """Insert data into the replay buffer."""
    raise ValueError('This function should not be called.')

  def sample_internal(self, buffer_state: State) -> Tuple[State, Sample]:
    """Sample a batch of data."""
    raise ValueError('This function should not be called.')

  def size(self, buffer_state: State) -> int:
    axis_name = 'x'

    def psize(buffer_state):
      return jax.lax.psum(self._buffer.size(buffer_state), axis_name=axis_name)

    return jax.pmap(psize, axis_name=axis_name)(buffer_state)[0]


class PjitWrapper(ReplayBuffer[State, Sample]):
  """Wrapper to distribute the buffer on multiple devices with pjit.

  Each device stores a part of the replay buffer depending on its index on axis
  'axis_name'.
  The total capacity of this replay buffer is the size of the mesh multiplied
  by the size of the wrapped buffer.
  The sample size is also the size of the mesh multiplied by the size of the
  sample in the wrapped buffer. Sample batches from each shard are concatenated
  (i.e. for random sampling, each shard will sample from the data they can see).
  """

  def __init__(
      self,
      buffer: ReplayBuffer[State, Sample],
      mesh: jax.sharding.Mesh,
      axis_names: Sequence[str],
  ):
    """Constructor.

    Args:
      buffer: The buffer to replicate.
      mesh: Device mesh for pjitting context.
      axis_names: The axes along which the replay buffer data should be
        partitionned.
    """
    self._buffer = buffer
    self._mesh = mesh
    self._num_devices = math.prod(mesh.shape[name] for name in axis_names)

    def init(key: PRNGKey) -> State:
      keys = jax.random.split(key, self._num_devices)
      return jax.vmap(self._buffer.init)(keys)

    def insert(buffer_state: State, samples: Sample) -> State:
      samples = jax.tree_util.tree_map(
          lambda x: jnp.reshape(x, (-1, self._num_devices) + x.shape[1:]),
          samples,
      )
      # This is to enforce we're gonna iterate on the start of the batch before
      # the end of the batch.
      samples = jax.tree_util.tree_map(lambda x: jnp.swapaxes(x, 0, 1), samples)
      return jax.vmap(self._buffer.insert_internal)(buffer_state, samples)

    def sample(buffer_state: State) -> Tuple[State, Sample]:
      buffer_state, samples = jax.vmap(self._buffer.sample_internal)(
          buffer_state
      )
      samples = jax.tree_util.tree_map(lambda x: jnp.swapaxes(x, 0, 1), samples)
      samples = jax.tree_util.tree_map(
          lambda x: jnp.reshape(x, (-1,) + x.shape[2:]), samples
      )
      return buffer_state, samples

    def size(buffer_state: State) -> int:
      return jnp.sum(jax.vmap(self._buffer.size)(buffer_state))  # pytype: disable=bad-return-type  # jnp-type

    partition_spec = jax.sharding.PartitionSpec((axis_names),)
    self._partitioned_init = pjit.pjit(init, out_shardings=partition_spec)
    self._partitioned_insert = pjit.pjit(
        insert,
        out_shardings=partition_spec,
    )
    self._partitioned_sample = pjit.pjit(
        sample,
        out_shardings=partition_spec,
    )
    # This will return the TOTAL size across all devices.
    self._partitioned_size = pjit.pjit(size, out_shardings=None)

  def init(self, key: PRNGKey) -> State:
    """See base class."""
    with self._mesh:
      return self._partitioned_init(key)

  def insert(self, buffer_state: State, samples: Sample) -> State:
    """See base class."""
    self._buffer.check_can_insert(buffer_state, samples, self._num_devices)
    with self._mesh:
      return self._partitioned_insert(buffer_state, samples)

  def sample(self, buffer_state: State) -> Tuple[State, Sample]:
    """See base class."""
    self._buffer.check_can_sample(buffer_state, self._num_devices)
    with self._mesh:
      return self._partitioned_sample(buffer_state)

  def size(self, buffer_state: State) -> int:
    """See base class. The total size (sum of all partitions) is returned."""
    with self._mesh:
      return self._partitioned_size(buffer_state)

  def insert_internal(self, buffer_state: State, samples: Sample) -> State:
    """Insert data into the replay buffer."""
    raise ValueError('This function should not be called.')

  def sample_internal(self, buffer_state: State) -> Tuple[State, Sample]:
    """Sample a batch of data."""
    raise ValueError('This function should not be called.')


@flax.struct.dataclass
class PrimitiveReplayBufferState(Generic[Sample]):
  """The state of the primitive replay buffer."""

  samples: Optional[Sample] = None



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrappers that expose config data to jax tracing."""

from typing import Callable, Dict, List, Tuple, TypeVar, Union

from brax.v1 import pytree
from brax.v1.envs import env as brax_env
from brax.v1.experimental.tracing.customize import TracedConfig
import brax.v1.jumpy as jp
import jax

Pytree = TypeVar('Pytree')


class DomainRandomizationWrapper(brax_env.Wrapper):
  """Wraps environment methods to allow vectorization over config state."""

  def __init__(self, env_fn: Callable[..., brax_env.Env],
               custom_tree: Dict[str, Union[List[Pytree], Pytree,
                                            brax_env.State, jp.ndarray]],
               custom_tree_in_axes: Tuple[Dict[str, Union[List[Pytree], Pytree,
                                                          int]]]):
    self.custom_tree = custom_tree
    self.custom_tree_in_axes = custom_tree_in_axes
    self.env_fn = pytree.register(env_fn)

    def init_env_fn(custom_tree):

      def post_process_fn(config):
        config = TracedConfig(config, custom_tree=custom_tree)
        return config

      env = self.env_fn(post_process_fn=post_process_fn)
      return env

    v_init = jax.vmap(init_env_fn, in_axes=custom_tree_in_axes)
    env = jax.jit(v_init)(custom_tree)

    super().__init__(env)

  def reset(self, rng):

    def reset_fn(custom_tree):

      def post_process_fn(config):
        config = TracedConfig(config, custom_tree=custom_tree)
        return config

      env = self.env_fn(post_process_fn=post_process_fn)
      return env.reset(custom_tree['rng'])

    self.custom_tree_in_axes[0]['rng'] = 0
    self.custom_tree['rng'] = rng

    return jax.vmap(
        reset_fn, in_axes=self.custom_tree_in_axes)(
            self.custom_tree)

  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:

    def step_fn(custom_tree):

      def post_process_fn(config):
        config = TracedConfig(config, custom_tree=custom_tree)
        return config

      env = self.env_fn(post_process_fn=post_process_fn)
      return env.step(custom_tree['state'], custom_tree['action'])

    self.custom_tree_in_axes[0]['state'] = jax.tree.map(
        lambda x: 0 if x.shape else None, state)
    self.custom_tree_in_axes[0]['action'] = 0
    self.custom_tree['state'] = state
    self.custom_tree['action'] = action

    return jax.vmap(step_fn, in_axes=self.custom_tree_in_axes)(self.custom_tree)

  @property
  def observation_size(self) -> int:
    env = self.env_fn()
    return env.observation_size

  @property
  def action_size(self) -> int:
    env = self.env_fn()
    return env.action_size



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the gym wrapper."""

from absl.testing import absltest
from brax import envs
from brax.envs.wrappers import gym
from brax.envs.wrappers import training
import numpy as np


class GymTest(absltest.TestCase):

  def test_action_space(self):
    """Tests the action space of the GymWrapper."""
    base_env = envs.create('pusher')
    env = gym.GymWrapper(base_env)
    np.testing.assert_array_equal(
        env.action_space.low, base_env.sys.actuator.ctrl_range[:, 0])
    np.testing.assert_array_equal(
        env.action_space.high, base_env.sys.actuator.ctrl_range[:, 1])


  def test_vector_action_space(self):
    """Tests the action space of the VectorGymWrapper."""
    base_env = envs.create('pusher')
    env = gym.VectorGymWrapper(training.VmapWrapper(base_env, batch_size=256))
    np.testing.assert_array_equal(
        env.action_space.low,
        np.tile(base_env.sys.actuator.ctrl_range[:, 0], [256, 1]))
    np.testing.assert_array_equal(
        env.action_space.high,
        np.tile(base_env.sys.actuator.ctrl_range[:, 1], [256, 1]))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrappers to convert brax envs to gym envs."""
from typing import ClassVar, Optional

from brax.envs.base import PipelineEnv
from brax.io import image
import gym
from gym import spaces
from gym.vector import utils
import jax
import numpy as np


class GymWrapper(gym.Env):
  """A wrapper that converts Brax Env to one that follows Gym API."""

  # Flag that prevents `gym.register` from misinterpreting the `_step` and
  # `_reset` as signs of a deprecated gym Env API.
  _gym_disable_underscore_compat: ClassVar[bool] = True

  def __init__(self,
               env: PipelineEnv,
               seed: int = 0,
               backend: Optional[str] = None):
    self._env = env
    self.metadata = {
        'render.modes': ['human', 'rgb_array'],
        'video.frames_per_second': 1 / self._env.dt
    }
    self.seed(seed)
    self.backend = backend
    self._state = None

    obs = np.inf * np.ones(self._env.observation_size, dtype='float32')
    self.observation_space = spaces.Box(-obs, obs, dtype='float32')

    action = jax.tree.map(np.array, self._env.sys.actuator.ctrl_range)
    self.action_space = spaces.Box(action[:, 0], action[:, 1], dtype='float32')

    def reset(key):
      key1, key2 = jax.random.split(key)
      state = self._env.reset(key2)
      return state, state.obs, key1

    self._reset = jax.jit(reset, backend=self.backend)

    def step(state, action):
      state = self._env.step(state, action)
      info = {**state.metrics, **state.info}
      return state, state.obs, state.reward, state.done, info

    self._step = jax.jit(step, backend=self.backend)

  def reset(self):
    self._state, obs, self._key = self._reset(self._key)
    # We return device arrays for pytorch users.
    return obs

  def step(self, action):
    self._state, obs, reward, done, info = self._step(self._state, action)
    # We return device arrays for pytorch users.
    return obs, reward, done, info

  def seed(self, seed: int = 0):
    self._key = jax.random.PRNGKey(seed)

  def render(self, mode='human'):
    if mode == 'rgb_array':
      sys, state = self._env.sys, self._state
      if state is None:
        raise RuntimeError('must call reset or step before rendering')
      return image.render_array(sys, state.pipeline_state, 256, 256)
    else:
      return super().render(mode=mode)  # just raise an exception


class VectorGymWrapper(gym.vector.VectorEnv):
  """A wrapper that converts batched Brax Env to one that follows Gym VectorEnv API."""

  # Flag that prevents `gym.register` from misinterpreting the `_step` and
  # `_reset` as signs of a deprecated gym Env API.
  _gym_disable_underscore_compat: ClassVar[bool] = True

  def __init__(self,
               env: PipelineEnv,
               seed: int = 0,
               backend: Optional[str] = None):
    self._env = env
    self.metadata = {
        'render.modes': ['human', 'rgb_array'],
        'video.frames_per_second': 1 / self._env.dt
    }
    if not hasattr(self._env, 'batch_size'):
      raise ValueError('underlying env must be batched')

    self.num_envs = self._env.batch_size
    self.seed(seed)
    self.backend = backend
    self._state = None

    obs = np.inf * np.ones(self._env.observation_size, dtype='float32')
    obs_space = spaces.Box(-obs, obs, dtype='float32')
    self.observation_space = utils.batch_space(obs_space, self.num_envs)

    action = jax.tree.map(np.array, self._env.sys.actuator.ctrl_range)
    action_space = spaces.Box(action[:, 0], action[:, 1], dtype='float32')
    self.action_space = utils.batch_space(action_space, self.num_envs)

    def reset(key):
      key1, key2 = jax.random.split(key)
      state = self._env.reset(key2)
      return state, state.obs, key1

    self._reset = jax.jit(reset, backend=self.backend)

    def step(state, action):
      state = self._env.step(state, action)
      info = {**state.metrics, **state.info}
      return state, state.obs, state.reward, state.done, info

    self._step = jax.jit(step, backend=self.backend)

  def reset(self):
    self._state, obs, self._key = self._reset(self._key)
    return obs

  def step(self, action):
    self._state, obs, reward, done, info = self._step(self._state, action)
    return obs, reward, done, info

  def seed(self, seed: int = 0):
    self._key = jax.random.PRNGKey(seed)

  def render(self, mode='human'):
    if mode == 'rgb_array':
      sys, state = self._env.sys, self._state
      if state is None:
        raise RuntimeError('must call reset or step before rendering')
      return image.render_array(sys, state.pipeline_state.take(0), 256, 256)
    else:
      return super().render(mode=mode)  # just raise an exception



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Single primitive object, e.g. block, capsule, sphere."""
from brax.v1.experimental.composer import component_editor

ROOT = 'object'


def get_specs(size: float = 0.25,
              collider_type: str = 'sphere',
              no_obs: bool = False):
  """Get system config."""
  if collider_type == 'sphere':
    assert isinstance(size, (float, int)), size
    collider = dict(sphere=dict(radius=size))
    scale = size
  elif collider_type == 'capsule':
    assert len(size) == 2, size
    collider = dict(capsule=dict(radius=size[0], length=size[1]))
    scale = size[0]
  else:
    raise NotImplementedError(collider_type)
  config_json = dict(bodies=[
      dict(
          name=ROOT,
          colliders=[collider],
          inertia=dict(x=1.0, y=1.0, z=1.0),
          mass=1.0 * scale**3)
  ])
  message_str = component_editor.json2message_str(config_json)
  return dict(
      message_str=message_str,
      collides=(ROOT,),
      root=ROOT,
      term_fn=None,
      observers=('qp',) if not no_obs else ())



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for json."""

import json

from absl.testing import absltest
from brax import test_utils
from brax.generalized import pipeline
from brax.io import json as bjson
import jax
import jax.numpy as jp


class JsonTest(absltest.TestCase):

  def test_dumps(self):
    sys = test_utils.load_fixture('convex_convex.xml')
    state = pipeline.init(sys, sys.init_q, jp.zeros(sys.qd_size()))
    res = bjson.dumps(sys, [state])
    res = json.loads(res)

    self.assertIsInstance(res['geoms'], dict)
    self.assertSequenceEqual(
        sorted(res['geoms'].keys()),
        ['box', 'dodecahedron', 'pyramid', 'tetrahedron', 'world'],
    )
    self.assertLen(res['geoms']['world'], 1)

    for f in ['size', 'rgba', 'name', 'link_idx', 'pos', 'rot']:
      self.assertIn(f, res['geoms']['box'][0])

  def test_dumps_invalidstate_raises(self):
    sys = test_utils.load_fixture('convex_convex.xml')
    state = pipeline.init(sys, sys.init_q, jp.zeros(sys.qd_size()))
    state = jax.tree.map(lambda x: jp.stack([x, x]), state)
    with self.assertRaises(RuntimeError):
      bjson.dumps(sys, [state])


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BIG-Gym: crowd-sourced environments and behaviors."""
# pylint:disable=protected-access
# pylint:disable=g-complex-comprehension
import difflib
import functools
import importlib
import inspect
import itertools
from typing import Any, Union, Dict, Callable, Optional
from brax.v1 import envs as brax_envs
from brax.v1.envs import Env
from brax.v1.envs import wrappers
from brax.v1.experimental.biggym import registry
from brax.v1.experimental.biggym import tasks
from brax.v1.experimental.braxlines.envs import obs_indices
from brax.v1.experimental.composer import components as composer_components
from brax.v1.experimental.composer import composer
from brax.v1.experimental.composer import envs as composer_envs

is_multiagent = composer.is_multiagent
unwrap = composer.unwrap

ROOT_PATH = 'brax.experimental.biggym.registry'
ENVS = {}
REGISTRIES = {}
OPEN_ENDED_TRACKS = ('rl', 'mimax')
GOAL_ORIENTED_TRACKS = sorted(tasks.TASKS)
ENVS_BY_TRACKS = dict(
    open_ended={k: () for k in OPEN_ENDED_TRACKS},
    goal_oriented={k: () for k in GOAL_ORIENTED_TRACKS},
    goal_oriented_matches={
        k: () for k in tasks.SYMMETRIC_MA_TASKS + tasks.ASYMMETRIC_MA_TASKS
    },
)
COMPONENTS_BY_TRACKS = {k: () for k in tasks.TASKS}


def inspect_env(env_name: str):
  """Inspect env_params of an env (ComposerEnv only)."""
  assert_exists(env_name)
  if composer_envs.exists(env_name):
    return composer_envs.inspect_env(env_name)
  else:
    return {}, False


def assert_env_params(env_name: str,
                      env_params: Dict[str, Any],
                      ignore_kwargs: bool = True):
  """Inspect env_params of an env (ComposerEnv only)."""
  assert_exists(env_name)
  if composer_envs.exists(env_name):
    composer_envs.assert_env_params(env_name, env_params, ignore_kwargs)
  else:
    assert not env_params, env_params


def list_env():
  """List registered environments."""
  return sorted(brax_envs._envs) + composer_envs.list_env() + sorted(ENVS)


def exists(env_name: str):
  """If environment is registered."""
  return env_name in list_env()


def assert_exists(env_name: str):
  """Assert if an environment is registered."""
  exists_ = exists(env_name)
  if not exists_:
    closest = difflib.get_close_matches(env_name, list_env(), n=3)
    assert 0, f'{env_name} not found. Closest={closest}'


def get_func_kwargs(func):
  """Get keyword args of a function."""
  # first, unwrap functools.partial. only extra keyword arguments.
  partial_supported_params = {}
  while isinstance(func, functools.partial):
    partial_supported_params.update(func.keywords)
    func = func.func
  # secondly, inspect the original function for keyword arguments.
  fn_params = inspect.signature(func).parameters
  support_kwargs = any(
      v.kind == inspect.Parameter.VAR_KEYWORD for v in fn_params.values())
  supported_params = {
      k: v.default
      for k, v in fn_params.items()
      if v.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD and
      v.default != inspect._empty
  }
  supported_params.update(partial_supported_params)
  return supported_params, support_kwargs


def register_all(verbose: bool = False, **kwargs):
  """Register all registries."""
  for registry_name in registry.REGISTRIES:
    env_names, comp_names, task_env_names, _ = register(registry_name, **kwargs)
    if verbose:
      print((f'Registered {registry_name}: '
             f'{len(env_names)} envs, '
             f'{len(comp_names)} comps, '
             f'{len(task_env_names)} task_envs, '))


def register(registry_name: str,
             assert_override: bool = True,
             assert_format: bool = True,
             optional: bool = True):
  """Register all envs and components."""
  global ENVS, REGISTRIES, ENVS_BY_TRACKS

  assert (optional or registry_name not in REGISTRIES
         ), f'non-optional register() conflicts: {registry_name}'
  if registry_name in REGISTRIES:
    return REGISTRIES[registry_name]

  lib = importlib.import_module(f'{ROOT_PATH}.{registry_name}')
  metadata = {}
  if assert_format:  # assert submission format
    for attr in ('AUTHORS', 'CONTACTS', 'AFFILIATIONS', 'DESCRIPTIONS'):
      assert hasattr(lib, attr), f'{attr} must be defined as a tuple of strs'
      values = getattr(lib, attr)
      assert isinstance(
          values, tuple), f'{attr}={values} must be defined as a tuple of strs'
      assert all(isinstance(v, str) for v in values
                ), f'{attr}={values} must be defined as a tuple of strs'
      metadata[attr] = values

  envs = lib.ENVS or {}
  components = lib.COMPONENTS or {}
  envs = {registry.get_env_name(registry_name, k): v for k, v in envs.items()}
  components = {
      registry.get_comp_name(registry_name, k): v
      for k, v in components.items()
  }
  task_envs = []

  # register environments
  for env_name, env_info in sorted(envs.items()):
    if assert_override:
      assert not exists(env_name), f'{list_env()} contains {env_name}'
    assert 'module' in env_info, env_info
    env_module = env_info['module']
    if isinstance(env_module, str):  # e.g. jump_cheetah:JumpCheetah
      env_lib_path, env_mod_name = env_module.split(':', 1)
      env_lib = importlib.import_module(
          f'{ROOT_PATH}.{registry_name}.envs.{env_lib_path}')
      env_module = getattr(env_lib, env_mod_name)
    if composer_envs.is_env_desc(env_module):
      # register a ComposerEnv
      composer_envs.register_env(env_name, env_module, override=True)
    else:
      # register a standard Env
      ENVS[env_name] = env_module
    tracks = env_info.get('tracks', ['rl'])
    for track in tracks:
      assert track in OPEN_ENDED_TRACKS, f'{track} not in {OPEN_ENDED_TRACKS}'
      ENVS_BY_TRACKS['open_ended'][track] += (env_name,)
    if 'mimax' in tracks:
      # (MI-Max only) register obs_indices
      for indices_type, indices in env_info.get('obs_indices', {}).items():
        obs_indices.register_indices(env_name, indices_type, indices)

  # register components
  for comp_name, comp_info in sorted(components.items()):
    if assert_override:
      assert not composer_components.exists(
          comp_name
      ), f'{composer_components.list_components()} contains {comp_name}'
    comp_module = comp_info['module']
    comp_lib = composer_components.register_component(
        comp_name,
        load_path=f'{ROOT_PATH}.{registry_name}.components.{comp_module}',
        override=True)
    component_params = get_func_kwargs(comp_lib.get_specs)[0]
    for track in comp_info.get('tracks', []):
      assert (track
              in GOAL_ORIENTED_TRACKS), f'{track} not in {GOAL_ORIENTED_TRACKS}'
      track_env_name = tasks.get_task_env_name(track, comp_name)
      if assert_override:
        assert not exists(
            track_env_name), f'{list_env()} contains {track_env_name}'
      track_env_module = functools.partial(tasks.TASKS[track], comp_name,
                                           **component_params)
      # register a ComposerEnv
      composer_envs.register_env(
          track_env_name, track_env_module, override=True)
      task_envs += [track_env_name]
      ENVS_BY_TRACKS['goal_oriented'][track] += (track_env_name,)
      COMPONENTS_BY_TRACKS[track] += (comp_name,)

  assert envs or task_envs, 'no envs registered'
  REGISTRIES[registry_name] = (sorted(envs), sorted(components),
                               sorted(task_envs), metadata)
  return REGISTRIES[registry_name]


def register_match(track: str, comp1: str, comp2: str, assert_override=True):
  """Register a match."""
  comp1_lib = composer_components.register_component(comp1)
  comp1_params = get_func_kwargs(comp1_lib.get_specs)[0]
  comp2_lib = composer_components.register_component(comp2)
  comp2_params = get_func_kwargs(comp2_lib.get_specs)[0]
  track_env_name = tasks.get_match_env_name(track, comp1, comp2)
  if assert_override:
    assert not exists(track_env_name), f'{list_env()} contains {track_env_name}'
  track_env_module = functools.partial(
      tasks.TASKS[track],
      comp1,
      opponent=comp2,
      opponent_params=comp2_params,
      **comp1_params)
  # register a ComposerEnv
  composer_envs.register_env(track_env_name, track_env_module, override=True)
  ENVS_BY_TRACKS['goal_oriented_matches'][track] += (track_env_name,)
  return track_env_name


def register_matches(assert_override=True):
  """Register components by goal-oriented track to fight each other."""
  task_envs = []
  for track in tasks.SYMMETRIC_MA_TASKS:
    for comp1, comp2 in itertools.combinations(COMPONENTS_BY_TRACKS[track], 2):
      task_envs += [
          register_match(track, comp1, comp2, assert_override=assert_override)
      ]
  for track in tasks.ASYMMETRIC_MA_TASKS:
    for comp1, comp2 in itertools.product(COMPONENTS_BY_TRACKS[track],
                                          COMPONENTS_BY_TRACKS[track]):
      task_envs += [
          register_match(track, comp1, comp2, assert_override=assert_override)
      ]
  return task_envs


def create(env_name: str = None,
           episode_length: int = 1000,
           action_repeat: int = 1,
           auto_reset: bool = True,
           batch_size: Optional[int] = None,
           **kwargs) -> Env:
  """Creates an Env with a specified brax system."""
  assert_exists(env_name)
  if env_name in ENVS:
    env = ENVS[env_name](**kwargs)
    if episode_length is not None:
      env = wrappers.EpisodeWrapper(env, episode_length, action_repeat)
    if batch_size:
      env = wrappers.VectorWrapper(env, batch_size)
    if auto_reset:
      env = wrappers.AutoResetWrapper(env)
    return env
  else:
    return composer.create(
        env_name,
        episode_length=episode_length,
        action_repeat=action_repeat,
        auto_reset=auto_reset,
        batch_size=batch_size,
        **kwargs)


def create_fn(env_name: str = None, **kwargs) -> Callable[..., Env]:
  """Returns a function that when called, creates an Env."""
  return functools.partial(create, env_name=env_name, **kwargs)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Calculations for generating contacts."""

from typing import Optional
from brax import math
from brax.base import Contact
from brax.base import System
from brax.base import Transform
import jax
from jax import numpy as jp
from mujoco import mjx


def get(sys: System, x: Transform) -> Optional[Contact]:
  """Calculates contacts.

  Args:
    sys: system defining the kinematic tree and other properties
    x: link transforms in world frame

  Returns:
    Contact pytree
  """
  d = mjx.make_data(sys)
  if d.ncon == 0:
    return None

  @jax.vmap
  def local_to_global(pos1, quat1, pos2, quat2):
    pos = pos1 + math.rotate(pos2, quat1)
    mat = math.quat_to_3x3(math.quat_mul(quat1, quat2))
    return pos, mat

  x = x.concatenate(Transform.zero((1,)))
  xpos = x.pos[sys.geom_bodyid - 1]
  xquat = x.rot[sys.geom_bodyid - 1]
  geom_xpos, geom_xmat = local_to_global(
      xpos, xquat, sys.geom_pos, sys.geom_quat
  )

  # pytype: disable=wrong-arg-types
  d = d.replace(geom_xpos=geom_xpos, geom_xmat=geom_xmat)
  d = mjx.collision(sys, d)
  # pytype: enable=wrong-arg-types

  c = d.contact
  elasticity = (sys.elasticity[c.geom1] + sys.elasticity[c.geom2]) * 0.5

  body1 = jp.array(sys.geom_bodyid)[c.geom1] - 1
  body2 = jp.array(sys.geom_bodyid)[c.geom2] - 1
  link_idx = (body1, body2)

  return Contact(elasticity=elasticity, link_idx=link_idx, **c.__dict__)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ant."""
# pylint:disable=protected-access
import brax.v1 as brax
from brax.v1.envs import ant
from brax.v1.experimental.composer import component_editor
from brax.v1.experimental.composer.components import common

COLLIDES = ('$ Torso', '$ Body 4', '$ Body 7', '$ Body 10', '$ Body 13')

ROOT = '$ Torso'

DEFAULT_OBSERVERS = ('root_z_joints', 'cfrc')


def term_fn(done, sys, qp: brax.QP, info: brax.Info, component,
            **unused_kwargs):
  """Termination."""
  done = common.height_term_fn(
      done,
      sys,
      qp,
      info,
      component,
      max_height=1.0,
      min_height=0.2,
      **unused_kwargs)
  done = common.upright_term_fn(done, sys, qp, info, component, **unused_kwargs)
  return done


def get_specs():
  return dict(
      message_str=component_editor.filter_message_str(ant._SYSTEM_CONFIG_SPRING,
                                                      'Ground'),
      collides=COLLIDES,
      root=ROOT,
      term_fn=term_fn,
      observers=DEFAULT_OBSERVERS)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brax.base."""

from absl.testing import absltest
from brax import test_utils
import jax
import numpy as np


class BaseTest(absltest.TestCase):

  def test_write_mass_array(self):
    sys = test_utils.load_fixture('ant.xml')
    rng = jax.random.PRNGKey(0)
    noise = jax.random.uniform(rng, (sys.link.inertia.mass.shape[0],))
    new_mass = sys.link.inertia.mass + noise

    sys_w = sys.tree_replace({'link.inertia.mass': new_mass})
    np.testing.assert_array_equal(sys_w.link.inertia.mass, new_mass)

  def test_write_mass_value(self):
    sys = test_utils.load_fixture('ant.xml')
    sys_w = sys.tree_replace({'link.inertia.mass': 1.0})
    self.assertEqual(sys_w.link.inertia.mass, 1.0)

  def test_write_array(self):
    sys = test_utils.load_fixture('ant.xml')
    np.random.seed(0)

    expected = np.random.uniform(sys.elasticity.shape)
    sys_w = sys.tree_replace({'elasticity': expected})
    np.testing.assert_array_equal(sys_w.elasticity, expected)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to manipulate configuration objects.

E.g. configuration objects:
    dict(a=[2,3,4], b=1) --> sample: dict(a=2,b=1)
    [dict(a=[2,3,4], b=1), dict(a=1, b=[2,3])] --> sample: dict(a=1,b=2)

The first level can be a list (in which case, each item is expanded separately),
or a dictionary.
After the first level, all lists are expanded while the rest of structures
specified by dictionaries are kept the same.
"""
import copy
from typing import Dict, Tuple, Any
import numpy as np


def filter_configuration(config, root=True, include_keys=()):
  """Filter configuration (non-recursive)."""
  if isinstance(config, dict):
    ret = {k: v for k, v in config.items() if k in include_keys}
  elif root and isinstance(config, list):
    ret = [
        filter_configuration(c, root=False, include_keys=include_keys)
        for c in config
    ]
  else:
    ret = config
  return ret


def get_compressed_name_from_keys(config: Dict[str, Any],
                                  keys: Tuple[Tuple[str]],
                                  allow_missing: bool = True):
  """Generate a compressed name from keys wrt config."""
  assert not isinstance(config, list), config
  name = ''
  used_pre_keys_strs = []
  for pre_keys in keys:
    v = config
    pre_keys_str = ''
    missing = False
    for k in pre_keys:
      if allow_missing and k not in v:
        missing = True
        break
      v = v[k]
      k_str = ''.join([s[0] for s in k.split('_')])  # learning_rate -> lr
      pre_keys_str = f'{pre_keys_str}.{k_str}' if pre_keys_str else k_str
    if missing:
      continue
    if pre_keys_str in used_pre_keys_strs:
      i = 0
      while f'{pre_keys_str}{i}' in used_pre_keys_strs and i < 10:
        i += 1
      assert i < 10, 'too many conflicts'
      pre_keys_str = f'{pre_keys_str}{i}'
    used_pre_keys_strs.append(pre_keys_str)
    if isinstance(v, bool):
      v = str(v)[0]  # True/False -> 'T', 'F'
    elif v is None:
      v = str(v)[0]  # None -> 'N'
    else:
      v = str(v)
    pre_keys_str += '_' + v
    name = f'{name}__{pre_keys_str}' if name else pre_keys_str
  return name


def list_keys_to_expand(config, root=True, pre_keys=()):
  """List the keys corresponding to List or callable."""
  if isinstance(config, dict):
    keys = ()
    for k, v in sorted(config.items()):
      keys += list_keys_to_expand(v, root=False, pre_keys=pre_keys + (k,))
    return keys
  elif (not root and isinstance(config, list)) or callable(config):
    assert pre_keys
    return (pre_keys,)
  elif root and isinstance(config, list):
    return tuple(
        list_keys_to_expand(v, root=False, pre_keys=pre_keys) for v in config)
  else:
    return ()


def count_configuration(config, root=True, num_samples_per_dist=1):
  """Recursively count configuration."""
  count = 1
  if isinstance(config, dict):
    for _, v in sorted(config.items()):
      count *= count_configuration(
          v, root=False, num_samples_per_dist=num_samples_per_dist)
  elif callable(config):
    assert num_samples_per_dist > 0, ('callable not allowed in config with '
                                      'num_samples_per_dist < 1')
    count *= num_samples_per_dist
  elif isinstance(config, list):
    if root:
      count = ()
      for c in config:
        count += (count_configuration(
            c, root=False, num_samples_per_dist=num_samples_per_dist),)
    else:
      count *= len(config)
  return count


def index_configuration(config, index=0, root=True, count=(), return_copy=True):
  """Configuration."""
  if isinstance(config, dict):
    c = {}
    for k, v in sorted(config.items()):
      c_, index = index_configuration(v, index=index, root=False)
      c[k] = c_
    if root and return_copy:
      c = copy.deepcopy(c)
    return c, index
  elif callable(config):
    raise NotImplementedError(
        'callable not allowed; call sample_configuration_dist first')
  elif isinstance(config, list):
    if root:
      i = 0
      index_ = index
      while i < len(config) and index_ >= count[i]:
        index_ -= count[i]
        i += 1
      assert i < len(config) and index_ >= 0 and index_ <= count[
          i], f'invalid index={index} wrt count={count}'
      c, i = index_configuration(config[i], index=index_, root=False)
      if return_copy:
        c = copy.deepcopy(c)
      return c, i
    else:
      i = index % len(config)
      index_ = int(index / len(config))
      return config[i], index_
  else:
    assert not root
    return config, index


def sample_configuration_dist(config, root=True, num_samples_per_dist=1):
  """Expand configuration distribution specification."""
  if isinstance(config, dict):
    return {
        k: sample_configuration_dist(
            v, root=False, num_samples_per_dist=num_samples_per_dist)
        for k, v in sorted(config.items())
    }
  elif isinstance(config, list) and root:
    return [
        sample_configuration_dist(
            c, root=False, num_samples_per_dist=num_samples_per_dist)
        for c in config
    ]
  elif callable(config):
    return [config() for _ in range(num_samples_per_dist)]
  else:
    return config


def get_configuration_sample(config, root=True):
  """Get a sample of config."""
  if isinstance(config, dict):
    return {
        k: get_configuration_sample(v, root=False)
        for k, v in sorted(config.items())
    }
  elif isinstance(config, list):
    if root:
      return get_configuration_sample(
          config[np.random.randint(len(config))], root=False)
    else:
      return config[np.random.randint(len(config))]
  elif callable(config):
    return config()
  else:
    return config



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains a ur5e robot arm to move its end effector to a sequence of targets.

The 6 joints have been placed faithfully to the actual joint locations of the
ur5e robot arm model.  Because brax does not yet support meshes, the capsule
locations for collider boundaries are only approximate.

See https://www.universal-robots.com/products/ur5-robot/ for more details.
"""

from typing import Tuple

import brax.v1 as brax
from brax.v1 import jumpy as jp
from brax.v1 import math
from brax.v1.envs import env


class Ur5e(env.Env):
  """Trains a UR5E robotic arm to touch a sequence of random targets."""

  def __init__(self, legacy_spring=False, **kwargs):
    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
    super().__init__(config=config, **kwargs)
    self.target_idx = self.sys.body.index['Target']
    self.torso_idx = self.sys.body.index['wrist_3_link']
    self.target_radius = .02
    self.target_distance = .5

  def reset(self, rng: jp.ndarray) -> env.State:
    qp = self.sys.default_qp()
    rng, target = self._random_target(rng)
    pos = jp.index_update(qp.pos, self.target_idx, target)
    qp = qp.replace(pos=pos)
    info = self.sys.info(qp)
    obs = self._get_obs(qp, info)
    reward, done, zero = jp.zeros(3)
    metrics = {
        'hits': zero,
        'weightedHits': zero,
        'movingToTarget': zero,
    }
    info = {'rng': rng}
    return env.State(qp, obs, reward, done, metrics, info)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    qp, info = self.sys.step(state.qp, action)
    obs = self._get_obs(qp, info)

    # small reward for end effector moving towards target
    torso_delta = qp.pos[self.torso_idx] - state.qp.pos[self.torso_idx]
    target_rel = qp.pos[self.target_idx] - qp.pos[self.torso_idx]
    target_dist = jp.norm(target_rel)
    target_dir = target_rel / (1e-6 + target_dist)
    moving_to_target = .1 * jp.dot(torso_delta, target_dir)

    # big reward for reaching target
    target_hit = target_dist < self.target_radius
    target_hit = jp.where(target_hit, jp.float32(1), jp.float32(0))
    weighted_hit = target_hit

    reward = moving_to_target + weighted_hit

    state.metrics.update(
        hits=target_hit,
        weightedHits=weighted_hit,
        movingToTarget=moving_to_target,
    )

    # teleport any hit targets
    rng, target = self._random_target(state.info['rng'])
    target = jp.where(target_hit, target, qp.pos[self.target_idx])
    pos = jp.index_update(qp.pos, self.target_idx, target)
    qp = qp.replace(pos=pos)
    state.info.update(rng=rng)
    return state.replace(qp=qp, obs=obs, reward=reward)

  def _get_obs(self, qp: brax.QP, info: brax.Info) -> jp.ndarray:
    """Egocentric observation of target and arm body."""
    torso_fwd = math.rotate(jp.array([1., 0., 0.]), qp.rot[self.torso_idx])
    torso_up = math.rotate(jp.array([0., 0., 1.]), qp.rot[self.torso_idx])

    v_inv_rotate = jp.vmap(math.inv_rotate, include=(True, False))

    pos_local = qp.pos - qp.pos[self.torso_idx]
    pos_local = v_inv_rotate(pos_local, qp.rot[self.torso_idx])
    vel_local = v_inv_rotate(qp.vel, qp.rot[self.torso_idx])

    target_local = pos_local[self.target_idx]
    target_local_mag = jp.reshape(jp.norm(target_local), -1)
    target_local_dir = target_local / (1e-6 + target_local_mag)

    pos_local = jp.reshape(pos_local, -1)
    vel_local = jp.reshape(vel_local, -1)

    contact_mag = jp.sum(jp.square(info.contact.vel), axis=-1)
    contacts = jp.where(contact_mag > 0.00001, 1, 0)  # pytype: disable=wrong-arg-types  # jax-ndarray

    return jp.concatenate([
        torso_fwd, torso_up, target_local_mag, target_local_dir, pos_local,
        vel_local, contacts
    ])

  def _random_target(self, rng: jp.ndarray) -> Tuple[jp.ndarray, jp.ndarray]:
    """Returns a target location in a random circle slightly above xy plane."""
    rng, rng1, rng2 = jp.random_split(rng, 3)
    dist = self.target_radius + self.target_distance * jp.random_uniform(rng1)
    ang = jp.pi * 2. * jp.random_uniform(rng2)
    target_x = dist * jp.cos(ang)
    target_y = dist * jp.sin(ang)
    target_z = .5
    target = jp.array([target_x, target_y, target_z]).transpose()
    return rng, target


_SYSTEM_CONFIG = """
  bodies {
    name: "floor"
    colliders {
      plane {
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
    frozen {
      all: true
    }
  }
  bodies {
    name: "shoulder_link"
    colliders {
      position {
        y: 0.06682991981506348
      }
      rotation {
        x: 90.0
      }
      capsule {
        radius: 0.05945208668708801
        length: 0.13365983963012695
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "upper_arm_link"
    colliders {
      position {
        z: 0.21287038922309875
      }
      rotation {
      }
      capsule {
        radius: 0.05968618765473366
        length: 0.5446449518203735
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "forearm_link"
    colliders {
      position {
        z: 0.1851803958415985
      }
      rotation {
      }
      capsule {
        radius: 0.05584339052438736
        length: 0.48926496505737305
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "wrist_1_link"
    colliders {
      position {
        y: 0.10467606782913208
      }
      rotation {
        x: 90.0
      }
      capsule {
        radius: 0.038744933903217316
        length: 0.10467606782913208
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "wrist_2_link"
    colliders {
      position {
        z: 0.052344050258398056
      }
      rotation {
      }
      capsule {
        radius: 0.03879201412200928
        length: 0.10468810051679611
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "wrist_3_link"
    colliders {
      position {
        y: -0.04025782644748688
      }
      rotation {
        x: 90.0
      }
      capsule {
        radius: 0.01725015603005886
        length: 0.08051565289497375
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "Target"
    colliders {
      sphere {
        radius: 0.1
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
    frozen {
      all: true
    }
  }
  joints {
    name: "shoulder_pan_joint"
    parent: "floor"
    child: "shoulder_link"
    parent_offset {
      z: 0.163
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  joints {
    name: "shoulder_lift_joint"
    parent: "shoulder_link"
    child: "upper_arm_link"
    parent_offset {
      y: 0.138
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  joints {
    name: "elbow_joint"
    parent: "upper_arm_link"
    child: "forearm_link"
    parent_offset {
      y: -0.13
      z: 0.425
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  joints {
    name: "wrist_1_joint"
    parent: "forearm_link"
    child: "wrist_1_link"
    parent_offset {
      z: 0.3919999897480011
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  joints {
    name: "wrist_2_joint"
    parent: "wrist_1_link"
    child: "wrist_2_link"
    parent_offset {
      y: 0.12700000405311584
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  joints {
    name: "wrist_3_joint"
    parent: "wrist_2_link"
    child: "wrist_3_link"
    parent_offset {
      z: 0.1
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  actuators {
    name: "shoulder_pan_joint"
    joint: "shoulder_pan_joint"
    strength: 100.0
    torque {
    }
  }
  actuators {
    name: "shoulder_lift_joint"
    joint: "shoulder_lift_joint"
    strength: 100.0
    torque {
    }
  }
  actuators {
    name: "elbow_joint"
    joint: "elbow_joint"
    strength: 100.0
    torque {
    }
  }
  actuators {
    name: "wrist_1_joint"
    joint: "wrist_1_joint"
    strength: 100.0
    torque {
    }
  }
  actuators {
    name: "wrist_2_joint"
    joint: "wrist_2_joint"
    strength: 100.0
    torque {
    }
  }
  actuators {
    name: "wrist_3_joint"
    joint: "wrist_3_joint"
    strength: 100.0
    torque {
    }
  }
  gravity {
    z: -9.81
  }
  collide_include {}
  angular_damping: -0.05
  dt: 0.02
  substeps: 8
  dynamics_mode: "pbd"
  """

_SYSTEM_CONFIG_SPRING = """
  bodies {
    name: "floor"
    colliders {
      plane {
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
    frozen {
      all: true
    }
  }
  bodies {
    name: "shoulder_link"
    colliders {
      position {
        y: 0.06682991981506348
      }
      rotation {
        x: 90.0
      }
      capsule {
        radius: 0.05945208668708801
        length: 0.13365983963012695
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "upper_arm_link"
    colliders {
      position {
        z: 0.21287038922309875
      }
      rotation {
      }
      capsule {
        radius: 0.05968618765473366
        length: 0.5446449518203735
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "forearm_link"
    colliders {
      position {
        z: 0.1851803958415985
      }
      rotation {
      }
      capsule {
        radius: 0.05584339052438736
        length: 0.48926496505737305
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "wrist_1_link"
    colliders {
      position {
        y: 0.10467606782913208
      }
      rotation {
        x: 90.0
      }
      capsule {
        radius: 0.038744933903217316
        length: 0.10467606782913208
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "wrist_2_link"
    colliders {
      position {
        z: 0.052344050258398056
      }
      rotation {
      }
      capsule {
        radius: 0.03879201412200928
        length: 0.10468810051679611
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "wrist_3_link"
    colliders {
      position {
        y: -0.04025782644748688
      }
      rotation {
        x: 90.0
      }
      capsule {
        radius: 0.01725015603005886
        length: 0.08051565289497375
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "Target"
    colliders {
      sphere {
        radius: 0.1
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
    frozen {
      all: true
    }
  }
  joints {
    name: "shoulder_pan_joint"
    stiffness: 40000.0
    parent: "floor"
    child: "shoulder_link"
    parent_offset {
      z: 0.163
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  joints {
    name: "shoulder_lift_joint"
    stiffness: 40000.0
    parent: "shoulder_link"
    child: "upper_arm_link"
    parent_offset {
      y: 0.138
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  joints {
    name: "elbow_joint"
    stiffness: 40000.0
    parent: "upper_arm_link"
    child: "forearm_link"
    parent_offset {
      y: -0.13
      z: 0.425
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  joints {
    name: "wrist_1_joint"
    stiffness: 40000.0
    parent: "forearm_link"
    child: "wrist_1_link"
    parent_offset {
      z: 0.3919999897480011
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  joints {
    name: "wrist_2_joint"
    stiffness: 40000.0
    parent: "wrist_1_link"
    child: "wrist_2_link"
    parent_offset {
      y: 0.12700000405311584
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  joints {
    name: "wrist_3_joint"
    stiffness: 40000.0
    parent: "wrist_2_link"
    child: "wrist_3_link"
    parent_offset {
      z: 0.1
    }
    child_offset {
    }
    rotation {
      z: 90.0
    }
    angular_damping: 50.0
    angle_limit {
      min: -360.0
      max: 360.0
    }
  }
  actuators {
    name: "shoulder_pan_joint"
    joint: "shoulder_pan_joint"
    strength: 100.0
    torque {
    }
  }
  actuators {
    name: "shoulder_lift_joint"
    joint: "shoulder_lift_joint"
    strength: 100.0
    torque {
    }
  }
  actuators {
    name: "elbow_joint"
    joint: "elbow_joint"
    strength: 100.0
    torque {
    }
  }
  actuators {
    name: "wrist_1_joint"
    joint: "wrist_1_joint"
    strength: 100.0
    torque {
    }
  }
  actuators {
    name: "wrist_2_joint"
    joint: "wrist_2_joint"
    strength: 100.0
    torque {
    }
  }
  actuators {
    name: "wrist_3_joint"
    joint: "wrist_3_joint"
    strength: 100.0
    torque {
    }
  }
  gravity {
    z: -9.81
  }
  collide_include {}
  angular_damping: -0.05
  baumgarte_erp: 0.1
  dt: 0.02
  substeps: 8
  dynamics_mode: "legacy_spring"
  """



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrappers to convert brax envs to DM Env envs."""
from typing import Optional

from brax.envs.base import PipelineEnv
from brax.io import image
import dm_env
from dm_env import specs
import jax
from jax import numpy as jp
import numpy as np


class DmEnvWrapper(dm_env.Environment):
  """A wrapper that converts Brax Env to one that follows Dm Env API."""

  def __init__(self,
               env: PipelineEnv,
               seed: int = 0,
               backend: Optional[str] = None):
    self._env = env
    self.seed(seed)
    self.backend = backend
    self._state = None

    if hasattr(self._env, 'observation_spec'):
      self._observation_spec = self._env.observation_spec()
    else:
      obs_high = jp.inf * jp.ones(self._env.observation_size, dtype='float32')
      self._observation_spec = specs.BoundedArray((self._env.observation_size,),
                                                  minimum=-obs_high,
                                                  maximum=obs_high,
                                                  dtype='float32',
                                                  name='observation')

    if hasattr(self._env, 'action_spec'):
      self._action_spec = self._env.action_spec()
    else:
      action = jax.tree.map(np.array, self._env.sys.actuator.ctrl_range)
      self._action_spec = specs.BoundedArray((self._env.action_size,),
                                             minimum=action[:, 0],
                                             maximum=action[:, 1],
                                             dtype='float32',
                                             name='action')

    self._reward_spec = specs.Array(shape=(), dtype=jp.dtype('float32'), name='reward')
    self._discount_spec = specs.BoundedArray(
        shape=(), dtype='float32', minimum=0., maximum=1., name='discount')
    if hasattr(self._env, 'discount_spec'):
      self._discount_spec = self._env.discount_spec()

    def reset(key):
      key1, key2 = jax.random.split(key)
      state = self._env.reset(key2)
      return state, state.obs, key1

    self._reset = jax.jit(reset, backend=self.backend)

    def step(state, action):
      state = self._env.step(state, action)
      info = {**state.metrics, **state.info}
      return state, state.obs, state.reward, state.done, info

    self._step = jax.jit(step, backend=self.backend)

  def reset(self):
    self._state, obs, self._key = self._reset(self._key)
    return dm_env.TimeStep(
        step_type=dm_env.StepType.FIRST,
        reward=None,
        discount=jp.float32(1.),
        observation=obs)

  def step(self, action):
    self._state, obs, reward, done, info = self._step(self._state, action)
    del info
    return dm_env.TimeStep(
        step_type=dm_env.StepType.MID if not done else dm_env.StepType.LAST,
        reward=reward,
        discount=jp.float32(1.),
        observation=obs)

  def seed(self, seed: int = 0):
    self._key = jax.random.PRNGKey(seed)

  def observation_spec(self):
    return self._observation_spec

  def action_spec(self):
    return self._action_spec

  def reward_spec(self):
    return self._reward_spec

  def discount_spec(self):
    return self._discount_spec

  def render(self):
    sys, state = self._env.sys, self._state
    if state is None:
      raise RuntimeError('must call reset or step before rendering')
    return image.render_array(sys, state.pipeline_state, 256, 256)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Physics pipeline for fully articulated dynamics and collisiion."""
# pylint:disable=g-multiple-import
# pylint:disable=g-importing-member
from typing import Optional
from brax.base import Contact, Motion, System, Transform
from brax.mjx.base import State
import jax
from jax import numpy as jp
from mujoco import mjx


def _reformat_contact(sys: System, data: State) -> State:
  """Reformats the mjx.Contact into a brax.base.Contact."""
  if data.contact is None or data.ncon == 0:
    return data

  elasticity = jp.zeros(data.contact.pos.shape[0])
  body1 = jp.array(sys.geom_bodyid)[data.contact.geom1] - 1
  body2 = jp.array(sys.geom_bodyid)[data.contact.geom2] - 1
  link_idx = (body1, body2)
  data = data.replace(
      contact=Contact(
          link_idx=link_idx, elasticity=elasticity, **data.contact.__dict__
      )
  )
  return data


def init(
    sys: System,
    q: jax.Array,
    qd: jax.Array,
    act: Optional[jax.Array] = None,
    unused_debug: bool = False,
) -> State:
  """Initializes physics data.

  Args:
    sys: a brax System
    q: (q_size,) joint angle vector
    qd: (qd_size,) joint velocity vector
    act: actuator activations
    unused_debug: ignored

  Returns:
    data: initial physics data
  """

  data = mjx.make_data(sys)
  data = data.replace(qpos=q, qvel=qd)
  if act is not None:
    data = data.replace(act=act)

  data = mjx.forward(sys, data)

  q, qd = data.qpos, data.qvel
  x = Transform(pos=data.xpos[1:], rot=data.xquat[1:])
  cvel = Motion(vel=data.cvel[1:, 3:], ang=data.cvel[1:, :3])
  offset = data.xpos[1:, :] - data.subtree_com[sys.body_rootid[1:]]
  offset = Transform.create(pos=offset)
  xd = offset.vmap().do(cvel)

  data = _reformat_contact(sys, data)
  return State(q=q, qd=qd, x=x, xd=xd, **data.__dict__)


def step(
    sys: System, state: State, act: jax.Array, unused_debug: bool = False
) -> State:
  """Performs a single physics step using position-based dynamics.

  Resolves actuator forces, joints, and forces at acceleration level, and
  resolves collisions at velocity level with baumgarte stabilization.

  Args:
    sys: a brax System
    state: physics data prior to step
    act: (act_size,) actuator input vector
    unused_debug: ignored

  Returns:
    x: updated link transform in world frame
    xd: updated link motion in world frame
  """
  data = state.replace(ctrl=act)
  data = mjx.step(sys, data)

  q, qd = data.qpos, data.qvel
  x = Transform(pos=data.xpos[1:], rot=data.xquat[1:])
  cvel = Motion(vel=data.cvel[1:, 3:], ang=data.cvel[1:, :3])
  offset = data.xpos[1:, :] - data.subtree_com[sys.body_rootid[1:]]
  offset = Transform.create(pos=offset)
  xd = offset.vmap().do(cvel)

  data = _reformat_contact(sys, data)
  return data.replace(q=q, qd=qd, x=x, xd=xd)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Import top-level classes and functions here for encapsulation/clarity."""

__version__ = '0.1.2'

import warnings

from brax.v1.physics.base import Info
from brax.v1.physics.base import QP
from brax.v1.physics.config_pb2 import Config
from brax.v1.physics.system import System

warnings.warn(
    'brax.v1 is deprecated and will be removed in a future release.',
    DeprecationWarning)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for Brax's Replay buffers."""

from typing import List

from absl.testing import absltest
from absl.testing import parameterized
from brax.training import replay_buffers
import jax
from jax.experimental.multihost_utils import process_allgather
import jax.numpy as jnp
import numpy as np


def get_dummy_data():
  return {'a': jnp.zeros(()), 'b': jnp.zeros((5, 5))}


def get_dummy_batch(batch_size: int = 8):
  return {
      'a': jnp.arange(batch_size, dtype=jnp.float32),
      'b': jnp.reshape(
          jnp.arange(batch_size * 5 * 5, dtype=jnp.float32), (batch_size, 5, 5)
      ),
  }


AXIS_NAME = 'x'
EXTRA_AXIS = 'y'


def get_mesh():
  devices = jax.devices()
  return jax.sharding.Mesh(
      np.array(devices).reshape(2 if len(devices) % 2 == 0 else 1, -1),
      (AXIS_NAME, EXTRA_AXIS),
  )


def sorted_data(buffer_state: replay_buffers.ReplayBufferState) -> List[int]:
  """Returns the data of the buffer, rolled so set the current position to 0."""
  return jnp.sort(jnp.ravel(buffer_state.data)).tolist()


def no_wrap(buffer):
  return buffer


def pjit_wrap(buffer):
  return replay_buffers.PjitWrapper(
      buffer,
      mesh=get_mesh(),
      axis_names=(AXIS_NAME,),
  )


def pmap_wrap(buffer):
  buffer = replay_buffers.PmapWrapper(
      buffer, local_device_count=2 if len(jax.devices()) > 1 else 1
  )
  buffer.insert_internal = jax.jit(buffer.insert_internal)
  buffer.sample_internal = jax.jit(buffer.sample_internal)
  return buffer


def jit_wrap(buffer):
  buffer.insert_internal = jax.jit(buffer.insert_internal)
  buffer.sample_internal = jax.jit(buffer.sample_internal)
  return buffer


QUEUE_FACTORIES = [
    (replay_buffers.UniformSamplingQueue, no_wrap),
    (replay_buffers.UniformSamplingQueue, pjit_wrap),
    (replay_buffers.UniformSamplingQueue, pmap_wrap),
    (replay_buffers.UniformSamplingQueue, jit_wrap),
    (replay_buffers.Queue, no_wrap),
    (replay_buffers.Queue, pjit_wrap),
    (replay_buffers.Queue, pmap_wrap),
    (replay_buffers.Queue, jit_wrap),
]


WRAPPERS = [no_wrap, pjit_wrap, pmap_wrap, jit_wrap]


def assert_equal(obj, data, expected_data):
  data = process_allgather(data)
  obj.assertTrue(
      jnp.all(data == expected_data), f'Not equal: {data} and {expected_data}'
  )


class QueueReplayTest(parameterized.TestCase):
  """Tests for Replay Buffers."""

  @parameterized.parameters(QUEUE_FACTORIES)
  def testInsert(self, queue_cls, wrapper):
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    replay_buffer = wrapper(
        queue_cls(
            max_replay_size=16 // size_denominator,
            dummy_data_sample=get_dummy_data(),
            sample_batch_size=2 // size_denominator,
        )
    )
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    assert_equal(self, replay_buffer.size(buffer_state), 0)

    buffer_state = replay_buffer.insert(buffer_state, get_dummy_batch())
    assert_equal(self, replay_buffer.size(buffer_state), 8)

    # Hit the max replay_size.
    buffer_state = replay_buffer.insert(buffer_state, get_dummy_batch(16))
    assert_equal(self, replay_buffer.size(buffer_state), 16)

  @parameterized.parameters(QUEUE_FACTORIES)
  def testInvalidStateShape(self, queue_cls, wrapper) -> None:
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    max_replay_size = 16 // size_denominator
    replay_buffer = wrapper(
        queue_cls(
            max_replay_size=max_replay_size,
            dummy_data_sample=get_dummy_data(),
            sample_batch_size=2 // mesh.shape[AXIS_NAME],
        )
    )
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)

    # Make sure inserting in the buffer works wihtout crashing.
    replay_buffer.insert(buffer_state, get_dummy_batch())

    # Expect an exception if `buffer_state.data` was corrupted.
    invalid_state = buffer_state.replace(data=jnp.arange(10))
    with self.assertRaises(ValueError) as context_manager:
      replay_buffer.insert(invalid_state, get_dummy_batch())
    self.assertContainsSubsequence(str(context_manager.exception), 'shape')

    # Expect an exception if batch_size is larger than max_replay_size.
    if wrapper in [pjit_wrap, pmap_wrap] and len(jax.devices()) != 1:
      return
    replay_buffer.insert(buffer_state, get_dummy_batch(max_replay_size))
    with self.assertRaisesRegex(
        ValueError,
        (
            'Trying to insert a batch of samples larger than the maximum replay'
            ' size. num_samples: 18, max replay size 16'
        ),
    ) as context_manager:
      replay_buffer.insert(buffer_state, get_dummy_batch(max_replay_size + 2))

  @parameterized.parameters(QUEUE_FACTORIES)
  def testInsertRollingBatches(self, queue_cls, wrapper) -> None:
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    max_size = 6 // size_denominator
    replay_buffer = wrapper(
        queue_cls(
            max_replay_size=max_size, dummy_data_sample=0, sample_batch_size=2
        )
    )
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    assert_equal(self, replay_buffer.size(buffer_state), 0)
    assert_equal(self, sorted_data(buffer_state), [0, 0, 0, 0, 0, 0])

    buffer_state = replay_buffer.insert(buffer_state, jnp.array([1, 2, 3, 4]))
    assert_equal(self, replay_buffer.size(buffer_state), 4)
    assert_equal(self, sorted_data(buffer_state), [0, 0, 1, 2, 3, 4])

    buffer_state = replay_buffer.insert(buffer_state, jnp.array([5, 6, 7, 8]))
    assert_equal(self, replay_buffer.size(buffer_state), 6)
    assert_equal(self, sorted_data(buffer_state), [3, 4, 5, 6, 7, 8])

    buffer_state = replay_buffer.insert(
        buffer_state, jnp.array([7, 8, 9, 10, 11, 12])
    )
    assert_equal(self, replay_buffer.size(buffer_state), 6)
    assert_equal(self, sorted_data(buffer_state), [7, 8, 9, 10, 11, 12])

  @parameterized.parameters(QUEUE_FACTORIES)
  def testInsertRolling2DimBatches(self, queue_cls, wrapper) -> None:
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    max_size = 8
    mesh = get_mesh()
    replay_buffer = wrapper(
        queue_cls(
            max_replay_size=max_size // size_denominator,
            dummy_data_sample=jnp.zeros((2,)),
            sample_batch_size=2,
        )
    )
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    assert_equal(self, replay_buffer.size(buffer_state), 0)
    if len(jax.devices()) == 1:
      assert_equal(
          self,
          buffer_state.data,
          [[0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0]],
      )

    buffer_state = replay_buffer.insert(
        buffer_state,
        jnp.array([[1, 1], [2, 2], [3, 3], [4, 4]], dtype=jnp.float32),
    )
    assert_equal(self, replay_buffer.size(buffer_state), 4)
    if len(jax.devices()) == 1:
      assert_equal(
          self,
          buffer_state.data,
          [[1, 1], [2, 2], [3, 3], [4, 4], [0, 0], [0, 0], [0, 0], [0, 0]],
      )

    buffer_state = replay_buffer.insert(
        buffer_state,
        jnp.array(
            [[5, 5], [6, 6], [7, 7], [8, 8], [9, 9], [10, 10]],
            dtype=jnp.float32,
        ),
    )
    assert_equal(self, replay_buffer.size(buffer_state), 8)
    if len(jax.devices()) == 1:
      assert_equal(
          self,
          buffer_state.data,
          [[3, 3], [4, 4], [5, 5], [6, 6], [7, 7], [8, 8], [9, 9], [10, 10]],
      )

  @parameterized.parameters(WRAPPERS)
  def testUniformSamplingQueueSample(self, wrapper):
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    replay_buffer = wrapper(
        replay_buffers.UniformSamplingQueue(
            max_replay_size=10 // size_denominator,
            dummy_data_sample=get_dummy_data(),
            sample_batch_size=2 // size_denominator,
        )
    )
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    assert_equal(self, replay_buffer.size(buffer_state), 0)

    buffer_state = replay_buffer.insert(buffer_state, get_dummy_batch(8))
    assert_equal(self, replay_buffer.size(buffer_state), 8)

    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, replay_buffer.size(buffer_state), 8)
    assert_equal(self, samples['a'].shape, (2,))
    assert_equal(self, samples['b'].shape, (2, 5, 5))
    for sample in samples['b']:
      assert_equal(
          self, jnp.reshape(sample - sample[0, 0], (-1,)), range(5 * 5)
      )

  @parameterized.parameters(WRAPPERS)
  def testUniformSamplingQueueCyclicSample(self, wrapper):
    replay_buffer = wrapper(
        replay_buffers.UniformSamplingQueue(
            max_replay_size=10, dummy_data_sample=0, sample_batch_size=2
        )
    )
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)

    assert_equal(self, replay_buffer.size(buffer_state), 0)

    buffer_state = replay_buffer.insert(
        buffer_state, jnp.zeros(10, dtype=jnp.int32)
    )
    assert_equal(self, replay_buffer.size(buffer_state), 10)
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(self, buffer_state.insert_position, 10)
    else:
      assert_equal(self, buffer_state.insert_position, [5, 5])

  @parameterized.parameters(WRAPPERS)
  def testQueueSamplePyTree(self, wrapper):
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    replay_buffer = wrapper(
        replay_buffers.Queue(
            max_replay_size=16 // size_denominator,
            dummy_data_sample=get_dummy_data(),
            sample_batch_size=2 // size_denominator,
        ),
    )
    rng = jax.random.PRNGKey(0)
    buffer_state = replay_buffer.init(rng)
    assert_equal(self, replay_buffer.size(buffer_state), 0)

    buffer_state = replay_buffer.insert(buffer_state, get_dummy_batch())
    assert_equal(self, replay_buffer.size(buffer_state), 8)

    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, replay_buffer.size(buffer_state), 6)
    assert_equal(self, samples['a'].shape, (2,))
    assert_equal(self, samples['b'].shape, (2, 5, 5))
    for sample in samples['b']:
      assert_equal(
          self, jnp.reshape(sample - sample[0, 0], (-1,)), range(5 * 5)
      )

  @parameterized.parameters(WRAPPERS)
  def testQueueSample(self, wrapper):
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    mesh = get_mesh()
    replay_buffer = wrapper(
        replay_buffers.Queue(
            max_replay_size=10 // size_denominator,
            dummy_data_sample=0,
            sample_batch_size=4 // size_denominator,
        )
    )
    rng = jax.random.PRNGKey(0)

    buffer_state = replay_buffer.init(rng)

    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(4))
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(
          self,
          buffer_state.data,
          [[0], [1], [2], [3], [0], [0], [0], [0], [0], [0]],
      )
      assert_equal(self, buffer_state.insert_position, 4)
    else:
      assert_equal(
          self,
          buffer_state.data,
          [[[0], [2], [0], [0], [0]], [[1], [3], [0], [0], [0]]],
      )
      assert_equal(self, buffer_state.insert_position, [2, 2])
    assert_equal(self, replay_buffer.size(buffer_state), 4)

    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(4, 10))
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(
          self,
          buffer_state.data,
          [[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]],
      )
      assert_equal(self, buffer_state.insert_position, 10)
    else:
      assert_equal(
          self,
          buffer_state.data,
          [[[0], [2], [4], [6], [8]], [[1], [3], [5], [7], [9]]],
      )
      assert_equal(self, buffer_state.insert_position, [5, 5])
    assert_equal(self, replay_buffer.size(buffer_state), 10)

    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, [0, 1, 2, 3])
    assert_equal(self, replay_buffer.size(buffer_state), 6)

    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, [4, 5, 6, 7])
    assert_equal(self, replay_buffer.size(buffer_state), 2)

    with self.assertRaisesRegex(
        ValueError, 'Trying to sample 4 elements, but only 2 available.'
    ):
      buffer_state, samples = replay_buffer.sample(buffer_state)

    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(20, 24))
    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, [8, 9, 20, 21])
    assert_equal(self, replay_buffer.size(buffer_state), 2)

  @parameterized.parameters(WRAPPERS)
  def testCyclicQueueSample(self, wrapper):
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    mesh = get_mesh()
    replay_buffer = wrapper(
        replay_buffers.Queue(
            max_replay_size=10 // size_denominator,
            dummy_data_sample=0,
            sample_batch_size=4 // size_denominator,
            cyclic=True,
        )
    )
    rng = jax.random.PRNGKey(0)

    buffer_state = replay_buffer.init(rng)

    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(6))
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(
          self,
          buffer_state.data,
          [[0], [1], [2], [3], [4], [5], [0], [0], [0], [0]],
      )
      assert_equal(self, buffer_state.insert_position, 6)
    else:
      assert_equal(
          self,
          buffer_state.data,
          [[[0], [2], [4], [0], [0]], [[1], [3], [5], [0], [0]]],
      )
      assert_equal(self, buffer_state.insert_position, [3, 3])
    assert_equal(self, replay_buffer.size(buffer_state), 6)

    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, [0, 1, 2, 3])
    assert_equal(self, replay_buffer.size(buffer_state), 6)

    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, [4, 5, 0, 1])
    assert_equal(self, replay_buffer.size(buffer_state), 6)

    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(6, 10))
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(
          self,
          buffer_state.data,
          [[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]],
      )
      assert_equal(self, buffer_state.insert_position, 10)
      assert_equal(self, buffer_state.sample_position, 2)
    else:
      assert_equal(
          self,
          buffer_state.data,
          [[[0], [2], [4], [6], [8]], [[1], [3], [5], [7], [9]]],
      )
      assert_equal(self, buffer_state.insert_position, [5, 5])
      assert_equal(self, buffer_state.sample_position, [1, 1])
    assert_equal(self, replay_buffer.size(buffer_state), 10)

    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, [2, 3, 4, 5])
    assert_equal(self, replay_buffer.size(buffer_state), 10)

    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, [6, 7, 8, 9])
    assert_equal(self, replay_buffer.size(buffer_state), 10)

    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(20, 24))
    assert_equal(self, replay_buffer.size(buffer_state), 10)

    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, [4, 5, 6, 7])
    assert_equal(self, replay_buffer.size(buffer_state), 10)

    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, [8, 9, 20, 21])
    assert_equal(self, replay_buffer.size(buffer_state), 10)

    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, [22, 23, 4, 5])
    assert_equal(self, replay_buffer.size(buffer_state), 10)

  @parameterized.parameters(WRAPPERS)
  def testQueueInsertWhenFull(self, wrapper):
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    max_replay_size = 10 // size_denominator
    replay_buffer = wrapper(
        replay_buffers.Queue(
            max_replay_size=max_replay_size,
            dummy_data_sample=0,
            sample_batch_size=4 // size_denominator,
        ),
    )
    rng = jax.random.PRNGKey(0)

    buffer_state = replay_buffer.init(rng)

    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(10))
    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(10, 12))
    if len(jax.devices()) == 1:
      assert_equal(
          self,
          buffer_state.data,
          [[2], [3], [4], [5], [6], [7], [8], [9], [10], [11]],
      )
      assert_equal(self, buffer_state.insert_position, 10)
    assert_equal(self, replay_buffer.size(buffer_state), 10)

  @parameterized.parameters(WRAPPERS)
  def testQueueWrappedSample(self, wrapper):
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    replay_buffer = wrapper(
        replay_buffers.Queue(
            max_replay_size=10 // size_denominator,
            dummy_data_sample=0,
            sample_batch_size=8 // size_denominator,
        ),
    )
    rng = jax.random.PRNGKey(0)

    buffer_state = replay_buffer.init(rng)
    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(10))
    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(10, 16))
    if len(jax.devices()) == 1:
      assert_equal(
          self,
          buffer_state.data,
          [[6], [7], [8], [9], [10], [11], [12], [13], [14], [15]],
      )
    assert_equal(self, replay_buffer.size(buffer_state), 10)
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(self, buffer_state.insert_position, 10)
    else:
      assert_equal(self, buffer_state.insert_position, [5, 5])

    # This sample contains elements from both the beggining and the end of
    # the buffer.
    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, jnp.array([6, 7, 8, 9, 10, 11, 12, 13]))
    assert_equal(self, replay_buffer.size(buffer_state), 2)
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(self, buffer_state.insert_position, 10)
    else:
      assert_equal(self, buffer_state.insert_position, [5, 5])

  @parameterized.parameters(WRAPPERS)
  def testQueueBatchSizeEqualsMaxSize(self, wrapper):
    batch_size = 8
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    mesh = get_mesh()
    replay_buffer = wrapper(
        replay_buffers.Queue(
            max_replay_size=batch_size // size_denominator,
            dummy_data_sample=0,
            sample_batch_size=batch_size // size_denominator,
        )
    )
    rng = jax.random.PRNGKey(0)

    buffer_state = replay_buffer.init(rng)

    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(batch_size))
    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, range(batch_size))
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(self, buffer_state.sample_position, 8)
    else:
      assert_equal(self, buffer_state.sample_position, [4, 4])

    buffer_state = replay_buffer.insert(
        buffer_state, jnp.zeros(batch_size, dtype=jnp.int32)
    )
    buffer_state = replay_buffer.insert(
        buffer_state, jnp.ones(batch_size, dtype=jnp.int32)
    )
    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, [1] * batch_size)
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(self, buffer_state.sample_position, 8)
    else:
      assert_equal(self, buffer_state.sample_position, [4, 4])

  @parameterized.parameters(WRAPPERS)
  def testQueueSampleFromEmpty(self, wrapper) -> None:
    batch_size = 10
    mesh = get_mesh()
    size_denominator = (
        1 if wrapper in [no_wrap, jit_wrap] else mesh.shape[AXIS_NAME]
    )
    replay_buffer = wrapper(
        replay_buffers.Queue(
            max_replay_size=batch_size // size_denominator,
            dummy_data_sample=0,
            sample_batch_size=batch_size // size_denominator,
        )
    )
    rng = jax.random.PRNGKey(0)

    buffer_state = replay_buffer.init(rng)

    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(batch_size))
    buffer_state, samples = replay_buffer.sample(buffer_state)
    assert_equal(self, samples, range(batch_size))
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(self, buffer_state.sample_position, 10)
    else:
      assert_equal(self, buffer_state.sample_position, [5, 5])

    with self.assertRaisesRegex(
        ValueError, 'Trying to sample 10 elements, but only 0 available.'
    ):
      replay_buffer.sample(buffer_state)
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(self, buffer_state.sample_position, 10)
    else:
      assert_equal(self, buffer_state.sample_position, [5, 5])

    buffer_state = replay_buffer.insert(buffer_state, jnp.arange(10, 14))
    if jax.device_count() == 1 or wrapper in [no_wrap, jit_wrap]:
      assert_equal(self, buffer_state.sample_position, 6)
    else:
      assert_equal(self, buffer_state.sample_position, [3, 3])
    with self.assertRaisesRegex(
        ValueError, 'Trying to sample 10 elements, but only 4 available.'
    ):
      buffer_state, samples = replay_buffer.sample(buffer_state)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Environment wrappers."""
import functools
from brax.v1.envs import env as brax_env
from jax import numpy as jnp


class ExtraStepArgsWrapper(brax_env.Wrapper):
  """Add optional kwargs to step() in Brax env."""

  def step(self, state: brax_env.State, action: jnp.ndarray, **kwargs):
    self.unwrapped.step = functools.partial(self.unwrapped.step, **kwargs)
    return self.env.step(state, action)



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Train VGCRL."""
import copy
import datetime
import functools
from typing import Dict, Any
from brax.v1.experimental.braxlines import experiments
from brax.v1.experimental.braxlines.common import logger_utils
from brax.v1.experimental.braxlines.envs.obs_indices import OBS_INDICES
from brax.v1.experimental.braxlines.training import ppo
from brax.v1.experimental.braxlines.vgcrl import evaluators as vgcrl_evaluators
from brax.v1.experimental.braxlines.vgcrl import utils as vgcrl_utils
from brax.v1.experimental.composer import composer
import jax
import tensorflow_probability.substrates.jax as tfp

tfd = tfp.distributions

TASK_KEYS = (('env_name',), ('obs_indices',), ('obs_scale',))


def train(train_job_params: Dict[str, Any],
          output_dir: str,
          return_dict: Dict[str, float] = None,
          progress_dict: Dict[str, float] = None,
          env_tag: str = None):
  """Train."""
  del env_tag
  return_dict = return_dict or {}
  progress_dict = progress_dict or {}
  logger_utils.save_config(
      f'{output_dir}/config.txt', train_job_params, verbose=True)
  config = copy.deepcopy(train_job_params)

  # @markdown **Task Parameters**
  env_name = config.pop('env_name', 'ant')
  obs_indices = config.pop('obs_indices', 'vel')
  obs_scale = config.pop('obs_scale', 5.0)
  obs_indices = OBS_INDICES[obs_indices][env_name]

  # @markdown **Experiment Parameters**
  algo_name = config.pop('algo_name', 'diayn')
  logits_clip_range = config.pop('logits_clip_range', 5.0)
  env_reward_multiplier = config.pop('env_reward_multiplier', 0.0)
  normalize_obs_for_disc = config.pop('normalize_obs_for_disc', False)
  seed = config.pop('seed', 0)
  evaluate_mi = config.pop('evaluate_mi', False)
  evaluate_lgr = config.pop('evaluate_lgr', False)
  diayn_num_skills = config.pop('diayn_num_skills', 8)
  disc_update_ratio = config.pop('disc_update_ratio', 1.0)
  spectral_norm = config.pop('spectral_norm', False)
  ppo_params = experiments.defaults.get_ppo_params(env_name, default='ant')
  ppo_params.update(config.pop('ppo_params', {}))
  assert not config, f'unused config: {config}'

  # @title Visualizing Brax environments
  # Create baseline environment to get observation specs
  base_env_fn = composer.create_fn(env_name=env_name)
  base_env = base_env_fn()
  env_obs_size = base_env.observation_size

  # Create discriminator-parameterized environment
  disc = vgcrl_utils.create_disc_fn(
      algo_name=algo_name,
      observation_size=env_obs_size,
      obs_indices=obs_indices,
      scale=obs_scale,
      diayn_num_skills=diayn_num_skills,
      logits_clip_range=logits_clip_range,
      spectral_norm=spectral_norm,
      env=base_env,
      normalize_obs=normalize_obs_for_disc)()
  extra_params = disc.init_model(rng=jax.random.PRNGKey(seed=seed))
  env_fn = vgcrl_utils.create_fn(
      env_name=env_name,
      wrapper_params=dict(
          env_reward_multiplier=env_reward_multiplier, disc=disc))
  eval_env_fn = functools.partial(env_fn, auto_reset=False)

  # make inference function and test goals
  core_env = env_fn()
  inference_fn = ppo.make_inference_fn(
      core_env.observation_size,
      core_env.action_size,
      normalize_observations=ppo_params.get('normalize_observation', True),
      extra_params=extra_params)
  inference_fn = jax.jit(inference_fn)
  goals = tfd.Uniform(
      low=-disc.obs_scale, high=disc.obs_scale).sample(
          seed=jax.random.PRNGKey(0), sample_shape=(10,))

  # @title Training
  tab = logger_utils.Tabulator(
      output_path=f'{output_dir}/training_curves.csv', append=False)

  # We determined some reasonable hyperparameters offline and share them here.
  train_fn = functools.partial(ppo.train, **ppo_params)

  times = [datetime.datetime.now()]
  plotpatterns = [
      'eval/episode_reward', 'losses/disc_loss', 'metrics/lgr',
      'metrics/entropy_all_', 'metrics/entropy_z_', 'metrics/mi_'
  ]

  def update_metrics_fn(num_steps, metrics, params):
    del num_steps
    if evaluate_mi:
      metrics.update(
          vgcrl_evaluators.estimate_empowerment_metric(
              env_fn=eval_env_fn,
              disc=disc,
              inference_fn=inference_fn,
              params=params,
              num_z=10,
              num_samples_per_z=10,
              time_subsampling=1,
              time_last_n=500,
              num_1d_bins=1000,
              num_2d_bins=30,
              verbose=True,
              seed=0))
    if evaluate_lgr:
      metrics.update(
          vgcrl_evaluators.estimate_latent_goal_reaching_metric(
              params=params,
              env_fn=eval_env_fn,
              disc=disc,
              inference_fn=inference_fn,
              goals=goals,
              num_samples_per_z=10,
              time_subsampling=1,
              time_last_n=500,
              seed=0))

  progress, plot, _, _ = experiments.get_progress_fn(
      plotpatterns,
      times,
      tab=tab,
      max_ncols=5,
      return_dict=return_dict,
      progress_dict=progress_dict,
      xlim=[0, train_fn.keywords['num_timesteps']],
      update_metrics_fn=update_metrics_fn)

  extra_loss_fns = dict(disc_loss=disc.disc_loss_fn) if extra_params else None
  extra_loss_update_ratios = dict(
      disc_loss=disc_update_ratio) if extra_params else None
  inference_fn, params, _ = train_fn(
      environment_fn=env_fn,
      seed=seed,
      progress_fn=progress,
      extra_params=extra_params,
      extra_loss_fns=extra_loss_fns,
      extra_loss_update_ratios=extra_loss_update_ratios,
  )
  plot(output_path=output_dir)

  return_dict.update(dict(time_to_jit=times[1] - times[0]))
  return_dict.update(dict(time_to_train=times[-1] - times[1]))
  print(f'time to jit: {times[1] - times[0]}')
  print(f'time to train: {times[-1] - times[1]}')

  vgcrl_evaluators.visualize_skills(
      env_fn=eval_env_fn,
      inference_fn=inference_fn,
      disc=disc,
      params=params,
      output_path=output_dir,
      verbose=True,
      num_z=20,
      num_samples_per_z=5,
      time_subsampling=10,
      time_last_n=500,
      seed=seed,
      save_video=True)

  return return_dict



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""multi-agent environments."""
import functools
import itertools
from typing import Any, Dict, Sequence
from brax.v1 import jumpy as jp
from brax.v1 import math
from brax.v1.experimental.composer import component_editor
from brax.v1.experimental.composer import reward_functions
from brax.v1.experimental.composer.composer_utils import merge_desc
from brax.v1.experimental.composer.observers import SimObserver as so
import numpy as np

MAX_DIST = 20
MIN_DIST = 0.5


def get_n_agents_desc(agents: Sequence[str],
                      agents_params: Sequence[str] = None,
                      init_r: float = 2):
  """Get n agents."""
  angles = np.linspace(0, 2 * np.pi, len(agents) + 1)
  agents_params = agents_params or ([None] * len(agents))
  components = {}
  edges = {}
  for i, (angle, agent,
          agent_params) in enumerate(zip(angles[:-1], agents, agents_params)):
    pos = (np.cos(angle) * init_r, np.sin(angle) * init_r, 0)
    components[f'agent{i}'] = dict(component=agent, pos=pos)
    if agent_params:
      components[f'agent{i}'].update(dict(component_params=agent_params))
  for k1, k2 in itertools.combinations(list(components), 2):
    if k1 == k2:
      continue
    k1, k2 = sorted([k1, k2])  # ensure the name is always sorted in order
    edge_name = component_editor.concat_comps(k1, k2)
    edges[edge_name] = dict(
        extra_observers=[dict(observer_type='root_vec', indices=(0, 1))])
  return dict(components=components, edges=edges)


def add_follow(env_desc: Dict[str, Any], leader_vel: float = 3.0):
  """Add follow task."""
  agent_groups = {}
  components = {}
  edges = {}
  agents = sorted(env_desc['components'])
  leader, followers = agents[0], agents[1:]
  # leader aims to run at a specific velocity
  components[leader] = dict(
      reward_fns=dict(
          goal=dict(
              reward_type='root_goal',
              sdcomp='vel',
              indices=(0, 1),
              offset=leader_vel + 2,
              target_goal=(leader_vel, 0))))
  agent_groups[leader] = dict(reward_agents=(leader,))
  # follower follows
  for agent in followers:
    edge_name = component_editor.concat_comps(agent, leader)
    edges[edge_name] = dict(
        reward_fns=dict(
            dist=dict(
                reward_type='root_dist', max_dist=MAX_DIST, offset=MAX_DIST +
                1)))
    agent_groups[agent] = dict(reward_names=(('dist', agent, leader),))
  merge_desc(
      env_desc,
      dict(agent_groups=agent_groups, components=components, edges=edges))
  return env_desc


def get_run_reward(scale: float = 1.0):
  return dict(
      reward_type=reward_functions.norm_reward,
      obs=lambda x: so('body', 'vel', x['root'], indices=(0, 1)),
      scale=-scale)


def add_chase(env_desc: Dict[str, Any]):
  """Add chase task."""
  agents = sorted(env_desc['components'])
  agent_groups = {agent: {'reward_names': ()} for agent in agents}
  components = {agent: {'reward_fns': {}} for agent in agents}
  edges = {}
  prey, predators = agents[0], agents[1:]
  for agent in predators:
    edge_name = component_editor.concat_comps(agent, prey)
    edges[edge_name] = dict(
        reward_fns=dict(
            # predators aim to chase the prey
            chase=dict(
                reward_type='root_dist',
                offset=MAX_DIST + 1,
                min_dist=MIN_DIST,
                done_bonus=1000 * MAX_DIST),
            # prey aims to run away from all predators
            escape=dict(
                reward_type='root_dist',
                scale=-1,
                max_dist=MAX_DIST,
                done_bonus=1000 * MAX_DIST,
            ),
        ))
    agent_groups[prey]['reward_names'] += (('escape', agent, prey),)
    agent_groups[agent]['reward_names'] += (('chase', agent, prey),)
  for agent in agents:
    # add velocity bonus for each agent
    components[agent]['reward_fns'].update(dict(run=get_run_reward()))
    agent_groups[agent]['reward_names'] += (('run', agent),)
  merge_desc(
      env_desc,
      dict(agent_groups=agent_groups, edges=edges, components=components))
  return env_desc


def get_ring_components(name: str = 'ring',
                        num_segments: int = 4,
                        radius: float = 3.0,
                        thickness: float = None,
                        offset: Sequence[float] = None):
  """Draw a ring with capsules."""
  offset = offset or [0, 0, 0]
  offset = jp.array(offset)
  thickness = thickness or radius / 40.
  components = {}
  angles = np.linspace(0, np.pi * 2, num_segments + 1)
  for i, angle in enumerate(angles[:-1]):
    k = f'{name}{i}'
    ring_length = radius * np.tan(np.pi / num_segments)
    components[k] = dict(
        component='singleton',
        component_params=dict(
            size=[thickness, ring_length * 2],
            collider_type='capsule',
            no_obs=True),
        pos=offset + jp.array(
            (radius * np.cos(angle), radius * np.sin(angle), -ring_length)),
        quat=math.euler_to_quat(jp.array([90, angle / jp.pi * 180, 0])),
        quat_origin=(0, 0, ring_length),
        frozen=True,
        collide=False)
  return components


def add_sumo(
    env_desc: Dict[str, Any],
    centering_scale: float = 1.,
    control_scale: float = 0.1,
    draw_scale: float = 0.,
    knocking_scale: float = 1.,
    opp_scale: float = 1.,
    ring_size: float = 3.,
    win_bonus: float = 1.,
):
  """Add a sumo task."""
  agents = sorted(env_desc['components'])
  agent_groups = {agent: {'reward_names': ()} for agent in agents}
  components = {agent: {'reward_fns': {}} for agent in agents}
  edges = {}
  yokozuna, komusubis = agents[0], agents[1:]
  for agent in komusubis:
    edge_name = component_editor.concat_comps(agent, yokozuna)
    edges[edge_name] = dict(
        reward_fns=dict(
            # komusubis wants to push out yokozuna
            komu_win_bonus=dict(
                reward_type=reward_functions.exp_norm_reward,
                obs=lambda x, y: so('body', 'pos', y['root'], indices=(0, 1)),
                max_dist=ring_size,
                done_bonus=win_bonus,
                scale=-knocking_scale,
            ),
            komu_lose_penalty=dict(
                reward_type=reward_functions.exp_norm_reward,
                obs=lambda x, y: so('body', 'pos', x['root'], indices=(0, 1)),
                max_dist=ring_size,
                done_bonus=-win_bonus,
                scale=centering_scale,
            ),
            # yokozuna wants to push out komusubis
            yoko_win_bonus=dict(
                reward_type=reward_functions.exp_norm_reward,
                obs=lambda x, y: so('body', 'pos', x['root'], indices=(0, 1)),
                max_dist=ring_size,
                done_bonus=win_bonus,
                scale=-knocking_scale,
            ),
            # each agent aims to be close to the center
            yoko_lose_penalty=dict(
                reward_type=reward_functions.exp_norm_reward,
                obs=lambda x, y: so('body', 'pos', y['root'], indices=(0, 1)),
                max_dist=ring_size,
                done_bonus=-win_bonus,
                scale=centering_scale,
            ),
            # move to opponent's direction
            komu_move_to_yoko=dict(
                reward_type=reward_functions.direction_reward,
                vel0=lambda x, y: so('body', 'vel', x['root'], indices=(0, 1)),
                vel1=lambda x, y: so('body', 'vel', y['root'], indices=(0, 1)),
                pos0=lambda x, y: so('body', 'pos', x['root'], indices=(0, 1)),
                pos1=lambda x, y: so('body', 'pos', y['root'], indices=(0, 1)),
                scale=opp_scale,
            ),
            yoko_move_to_komu=dict(
                reward_type=reward_functions.direction_reward,
                vel0=lambda x, y: so('body', 'vel', y['root'], indices=(0, 1)),
                vel1=lambda x, y: so('body', 'vel', x['root'], indices=(0, 1)),
                pos0=lambda x, y: so('body', 'pos', y['root'], indices=(0, 1)),
                pos1=lambda x, y: so('body', 'pos', x['root'], indices=(0, 1)),
                scale=opp_scale,
            ),
        ))
    agent_groups[agent]['reward_names'] += (('komu_win_bonus', agent, yokozuna),
                                            ('komu_lose_penalty', agent,
                                             yokozuna), ('komu_move_to_yoko',
                                                         agent, yokozuna))
    agent_groups[yokozuna]['reward_names'] += (('yoko_win_bonus', agent,
                                                yokozuna), ('yoko_lose_penalty',
                                                            agent, yokozuna),
                                               ('yoko_move_to_komu', yokozuna,
                                                agent))
  for agent in agents:
    components[agent]['reward_fns'].update(
        dict(
            control_penalty=dict(
                reward_type=reward_functions.control_reward,
                scale=control_scale,
            ),
            draw_penalty=dict(
                reward_type=reward_functions.constant_reward,
                value=-draw_scale,
            ),
        ))
    agent_groups[agent]['reward_names'] += (('control_penalty', agent),
                                            ('draw_penalty', agent))
  # add sumo ring
  components.update(get_ring_components(radius=ring_size, num_segments=20))
  merge_desc(
      env_desc,
      dict(agent_groups=agent_groups, edges=edges, components=components))
  return env_desc


def add_squidgame(env_desc: Dict[str, Any],
                  ring_size: float = 3.0,
                  run_scale: float = 0):
  """Add a simplified squid game task."""
  # TODO: finish reward functions
  agents = sorted(env_desc['components'])
  agent_groups = {agent: {'reward_names': ()} for agent in agents}
  components = {agent: {'reward_fns': {}} for agent in agents}
  edges = {}
  defender, attackers = agents[0], agents[1:]
  for agent in attackers:
    edge_name = component_editor.concat_comps(agent, defender)
    edges[edge_name] = dict(
        reward_fns=dict(
            # defenders aim to chase the attackers
            chase=dict(reward_type='root_dist', offset=2 * ring_size + 0.5),))
    agent_groups[defender]['reward_names'] += (('chase', agent, defender),)
  for agent in agents:
    if run_scale > 0:
      # add velocity bonus for each agent
      components[agent] = dict(
          reward_fns=dict(run=get_run_reward(scale=run_scale)))
      agent_groups[agent]['reward_names'] += (('run', agent),)
  # add rings
  components.update(
      get_ring_components(
          name='square',
          offset=(ring_size, 0, 0),
          radius=ring_size,
          thickness=ring_size / 40.,
          num_segments=4))
  components.update(
      get_ring_components(
          name='defender_circle',
          offset=(ring_size * 2, 0, 0),
          radius=ring_size / 5,
          thickness=ring_size / 40.,
          num_segments=10))
  components.update(
      get_ring_components(
          name='triangle',
          offset=(-ring_size / np.sqrt(3), 0, 0),
          radius=ring_size / np.sqrt(3),
          thickness=ring_size / 40.,
          num_segments=3))
  components.update(
      get_ring_components(
          name='attacker_circle',
          offset=(-ring_size * np.sqrt(3), 0, 0),
          radius=ring_size / 5,
          thickness=ring_size / 40.,
          num_segments=10))
  merge_desc(
      env_desc,
      dict(agent_groups=agent_groups, edges=edges, components=components))
  return env_desc


TASK_MAP = dict(
    follow=add_follow, chase=add_chase, sumo=add_sumo, squidgame=add_squidgame)


def create_desc(main_agent: str = 'ant',
                other_agent: str = 'ant',
                main_agent_params: Dict[str, Any] = None,
                other_agent_params: Dict[str, Any] = None,
                num_agents: int = 2,
                task: str = 'follow',
                init_r: float = 2.,
                **kwargs):
  """Creat env_desc."""
  if main_agent_params or other_agent_params:
    agents_params = [main_agent_params] + [other_agent_params] * (
        num_agents - 1)
  else:
    agents_params = None
  env_desc = get_n_agents_desc(
      agents=[main_agent] + [other_agent] * (num_agents - 1),
      agents_params=agents_params,
      init_r=init_r)

  return TASK_MAP[task](env_desc=env_desc, **kwargs)


ENV_DESCS = {k: functools.partial(create_desc, task=k) for k in TASK_MAP}



# Copyright 2024 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Proximal policy optimization training.

See: https://arxiv.org/pdf/1707.06347.pdf
"""

import functools
import time
from typing import Callable, Optional, Tuple, Union

from absl import logging
from brax import base
from brax import envs
from brax.training import acting
from brax.training import gradients
from brax.training import pmap
from brax.training import types
from brax.training.acme import running_statistics
from brax.training.acme import specs
from brax.training.agents.ppo import losses as ppo_losses
from brax.training.agents.ppo import networks as ppo_networks
from brax.training.types import Params
from brax.training.types import PRNGKey
from brax.v1 import envs as envs_v1
from etils import epath
import flax
import jax
import jax.numpy as jnp
import numpy as np
import optax
from orbax import checkpoint as ocp


InferenceParams = Tuple[running_statistics.NestedMeanStd, Params]
Metrics = types.Metrics

_PMAP_AXIS_NAME = 'i'


@flax.struct.dataclass
class TrainingState:
  """Contains training state for the learner."""
  optimizer_state: optax.OptState
  params: ppo_losses.PPONetworkParams
  normalizer_params: running_statistics.RunningStatisticsState
  env_steps: jnp.ndarray


def _unpmap(v):
  return jax.tree_util.tree_map(lambda x: x[0], v)


def _strip_weak_type(tree):
  # brax user code is sometimes ambiguous about weak_type.  in order to
  # avoid extra jit recompilations we strip all weak types from user input
  def f(leaf):
    leaf = jnp.asarray(leaf)
    return leaf.astype(leaf.dtype)
  return jax.tree_util.tree_map(f, tree)


def train(
    environment: Union[envs_v1.Env, envs.Env],
    num_timesteps: int,
    episode_length: int,
    action_repeat: int = 1,
    num_envs: int = 1,
    max_devices_per_host: Optional[int] = None,
    num_eval_envs: int = 128,
    learning_rate: float = 1e-4,
    entropy_cost: float = 1e-4,
    discounting: float = 0.9,
    seed: int = 0,
    unroll_length: int = 10,
    batch_size: int = 32,
    num_minibatches: int = 16,
    num_updates_per_batch: int = 2,
    num_evals: int = 1,
    num_resets_per_eval: int = 0,
    normalize_observations: bool = False,
    reward_scaling: float = 1.0,
    clipping_epsilon: float = 0.3,
    gae_lambda: float = 0.95,
    deterministic_eval: bool = False,
    network_factory: types.NetworkFactory[
        ppo_networks.PPONetworks
    ] = ppo_networks.make_ppo_networks,
    progress_fn: Callable[[int, Metrics], None] = lambda *args: None,
    normalize_advantage: bool = True,
    eval_env: Optional[envs.Env] = None,
    policy_params_fn: Callable[..., None] = lambda *args: None,
    randomization_fn: Optional[
        Callable[[base.System, jnp.ndarray], Tuple[base.System, base.System]]
    ] = None,
    restore_checkpoint_path: Optional[str] = None,
):
  """PPO training.

  Args:
    environment: the environment to train
    num_timesteps: the total number of environment steps to use during training
    episode_length: the length of an environment episode
    action_repeat: the number of timesteps to repeat an action
    num_envs: the number of parallel environments to use for rollouts
      NOTE: `num_envs` must be divisible by the total number of chips since each
        chip gets `num_envs // total_number_of_chips` environments to roll out
      NOTE: `batch_size * num_minibatches` must be divisible by `num_envs` since
        data generated by `num_envs` parallel envs gets used for gradient
        updates over `num_minibatches` of data, where each minibatch has a
        leading dimension of `batch_size`
    max_devices_per_host: maximum number of chips to use per host process
    num_eval_envs: the number of envs to use for evluation. Each env will run 1
      episode, and all envs run in parallel during eval.
    learning_rate: learning rate for ppo loss
    entropy_cost: entropy reward for ppo loss, higher values increase entropy of
      the policy
    discounting: discounting rate
    seed: random seed
    unroll_length: the number of timesteps to unroll in each environment. The
      PPO loss is computed over `unroll_length` timesteps
    batch_size: the batch size for each minibatch SGD step
    num_minibatches: the number of times to run the SGD step, each with a
      different minibatch with leading dimension of `batch_size`
    num_updates_per_batch: the number of times to run the gradient update over
      all minibatches before doing a new environment rollout
    num_evals: the number of evals to run during the entire training run.
      Increasing the number of evals increases total training time
    num_resets_per_eval: the number of environment resets to run between each
      eval. The environment resets occur on the host
    normalize_observations: whether to normalize observations
    reward_scaling: float scaling for reward
    clipping_epsilon: clipping epsilon for PPO loss
    gae_lambda: General advantage estimation lambda
    deterministic_eval: whether to run the eval with a deterministic policy
    network_factory: function that generates networks for policy and value
      functions
    progress_fn: a user-defined callback function for reporting/plotting metrics
    normalize_advantage: whether to normalize advantage estimate
    eval_env: an optional environment for eval only, defaults to `environment`
    policy_params_fn: a user-defined callback function that can be used for
      saving policy checkpoints
    randomization_fn: a user-defined callback function that generates randomized
      environments
    restore_checkpoint_path: the path used to restore previous model params

  Returns:
    Tuple of (make_policy function, network params, metrics)
  """
  assert batch_size * num_minibatches % num_envs == 0
  xt = time.time()

  process_count = jax.process_count()
  process_id = jax.process_index()
  local_device_count = jax.local_device_count()
  local_devices_to_use = local_device_count
  if max_devices_per_host:
    local_devices_to_use = min(local_devices_to_use, max_devices_per_host)
  logging.info(
      'Device count: %d, process count: %d (id %d), local device count: %d, '
      'devices to be used count: %d', jax.device_count(), process_count,
      process_id, local_device_count, local_devices_to_use)
  device_count = local_devices_to_use * process_count

  # The number of environment steps executed for every training step.
  env_step_per_training_step = (
      batch_size * unroll_length * num_minibatches * action_repeat)
  num_evals_after_init = max(num_evals - 1, 1)
  # The number of training_step calls per training_epoch call.
  # equals to ceil(num_timesteps / (num_evals * env_step_per_training_step *
  #                                 num_resets_per_eval))
  num_training_steps_per_epoch = np.ceil(
      num_timesteps
      / (
          num_evals_after_init
          * env_step_per_training_step
          * max(num_resets_per_eval, 1)
      )
  ).astype(int)

  key = jax.random.PRNGKey(seed)
  global_key, local_key = jax.random.split(key)
  del key
  local_key = jax.random.fold_in(local_key, process_id)
  local_key, key_env, eval_key = jax.random.split(local_key, 3)
  # key_networks should be global, so that networks are initialized the same
  # way for different processes.
  key_policy, key_value = jax.random.split(global_key)
  del global_key

  assert num_envs % device_count == 0

  v_randomization_fn = None
  if randomization_fn is not None:
    randomization_batch_size = num_envs // local_device_count
    # all devices gets the same randomization rng
    randomization_rng = jax.random.split(key_env, randomization_batch_size)
    v_randomization_fn = functools.partial(
        randomization_fn, rng=randomization_rng
    )

  if isinstance(environment, envs.Env):
    wrap_for_training = envs.training.wrap
  else:
    wrap_for_training = envs_v1.wrappers.wrap_for_training

  env = wrap_for_training(
      environment,
      episode_length=episode_length,
      action_repeat=action_repeat,
      randomization_fn=v_randomization_fn,
  )

  reset_fn = jax.jit(jax.vmap(env.reset))
  key_envs = jax.random.split(key_env, num_envs // process_count)
  key_envs = jnp.reshape(key_envs,
                         (local_devices_to_use, -1) + key_envs.shape[1:])
  env_state = reset_fn(key_envs)

  normalize = lambda x, y: x
  if normalize_observations:
    normalize = running_statistics.normalize
  ppo_network = network_factory(
      env_state.obs.shape[-1],
      env.action_size,
      preprocess_observations_fn=normalize)
  make_policy = ppo_networks.make_inference_fn(ppo_network)

  optimizer = optax.adam(learning_rate=learning_rate)

  loss_fn = functools.partial(
      ppo_losses.compute_ppo_loss,
      ppo_network=ppo_network,
      entropy_cost=entropy_cost,
      discounting=discounting,
      reward_scaling=reward_scaling,
      gae_lambda=gae_lambda,
      clipping_epsilon=clipping_epsilon,
      normalize_advantage=normalize_advantage)

  gradient_update_fn = gradients.gradient_update_fn(
      loss_fn, optimizer, pmap_axis_name=_PMAP_AXIS_NAME, has_aux=True)

  def minibatch_step(
      carry, data: types.Transition,
      normalizer_params: running_statistics.RunningStatisticsState):
    optimizer_state, params, key = carry
    key, key_loss = jax.random.split(key)
    (_, metrics), params, optimizer_state = gradient_update_fn(
        params,
        normalizer_params,
        data,
        key_loss,
        optimizer_state=optimizer_state)

    return (optimizer_state, params, key), metrics

  def sgd_step(carry, unused_t, data: types.Transition,
               normalizer_params: running_statistics.RunningStatisticsState):
    optimizer_state, params, key = carry
    key, key_perm, key_grad = jax.random.split(key, 3)

    def convert_data(x: jnp.ndarray):
      x = jax.random.permutation(key_perm, x)
      x = jnp.reshape(x, (num_minibatches, -1) + x.shape[1:])
      return x

    shuffled_data = jax.tree_util.tree_map(convert_data, data)
    (optimizer_state, params, _), metrics = jax.lax.scan(
        functools.partial(minibatch_step, normalizer_params=normalizer_params),
        (optimizer_state, params, key_grad),
        shuffled_data,
        length=num_minibatches)
    return (optimizer_state, params, key), metrics

  def training_step(
      carry: Tuple[TrainingState, envs.State, PRNGKey],
      unused_t) -> Tuple[Tuple[TrainingState, envs.State, PRNGKey], Metrics]:
    training_state, state, key = carry
    key_sgd, key_generate_unroll, new_key = jax.random.split(key, 3)

    policy = make_policy(
        (training_state.normalizer_params, training_state.params.policy))

    def f(carry, unused_t):
      current_state, current_key = carry
      current_key, next_key = jax.random.split(current_key)
      next_state, data = acting.generate_unroll(
          env,
          current_state,
          policy,
          current_key,
          unroll_length,
          extra_fields=('truncation',))
      return (next_state, next_key), data

    (state, _), data = jax.lax.scan(
        f, (state, key_generate_unroll), (),
        length=batch_size * num_minibatches // num_envs)
    # Have leading dimensions (batch_size * num_minibatches, unroll_length)
    data = jax.tree_util.tree_map(lambda x: jnp.swapaxes(x, 1, 2), data)
    data = jax.tree_util.tree_map(lambda x: jnp.reshape(x, (-1,) + x.shape[2:]),
                                  data)
    assert data.discount.shape[1:] == (unroll_length,)

    # Update normalization params and normalize observations.
    normalizer_params = running_statistics.update(
        training_state.normalizer_params,
        data.observation,
        pmap_axis_name=_PMAP_AXIS_NAME)

    (optimizer_state, params, _), metrics = jax.lax.scan(
        functools.partial(
            sgd_step, data=data, normalizer_params=normalizer_params),
        (training_state.optimizer_state, training_state.params, key_sgd), (),
        length=num_updates_per_batch)

    new_training_state = TrainingState(
        optimizer_state=optimizer_state,
        params=params,
        normalizer_params=normalizer_params,
        env_steps=training_state.env_steps + env_step_per_training_step)
    return (new_training_state, state, new_key), metrics

  def training_epoch(training_state: TrainingState, state: envs.State,
                     key: PRNGKey) -> Tuple[TrainingState, envs.State, Metrics]:
    (training_state, state, _), loss_metrics = jax.lax.scan(
        training_step, (training_state, state, key), (),
        length=num_training_steps_per_epoch)
    loss_metrics = jax.tree_util.tree_map(jnp.mean, loss_metrics)
    return training_state, state, loss_metrics

  training_epoch = jax.pmap(training_epoch, axis_name=_PMAP_AXIS_NAME)

  # Note that this is NOT a pure jittable method.
  def training_epoch_with_timing(
      training_state: TrainingState, env_state: envs.State,
      key: PRNGKey) -> Tuple[TrainingState, envs.State, Metrics]:
    nonlocal training_walltime
    t = time.time()
    training_state, env_state = _strip_weak_type((training_state, env_state))
    result = training_epoch(training_state, env_state, key)
    training_state, env_state, metrics = _strip_weak_type(result)

    metrics = jax.tree_util.tree_map(jnp.mean, metrics)
    jax.tree_util.tree_map(lambda x: x.block_until_ready(), metrics)

    epoch_training_time = time.time() - t
    training_walltime += epoch_training_time
    sps = (num_training_steps_per_epoch *
           env_step_per_training_step *
           max(num_resets_per_eval, 1)) / epoch_training_time
    metrics = {
        'training/sps': sps,
        'training/walltime': training_walltime,
        **{f'training/{name}': value for name, value in metrics.items()}
    }
    return training_state, env_state, metrics  # pytype: disable=bad-return-type  # py311-upgrade

  # Initialize model params and training state.
  init_params = ppo_losses.PPONetworkParams(
      policy=ppo_network.policy_network.init(key_policy),
      value=ppo_network.value_network.init(key_value),
  )

  training_state = TrainingState(  # pytype: disable=wrong-arg-types  # jax-ndarray
      optimizer_state=optimizer.init(init_params),  # pytype: disable=wrong-arg-types  # numpy-scalars
      params=init_params,
      normalizer_params=running_statistics.init_state(
          specs.Array(env_state.obs.shape[-1:], jnp.dtype('float32'))),
      env_steps=0)

  if num_timesteps == 0:
    return (
        make_policy,
        (training_state.normalizer_params, training_state.params),
        {},
    )

  if (
      restore_checkpoint_path is not None
      and epath.Path(restore_checkpoint_path).exists()
  ):
    logging.info('restoring from checkpoint %s', restore_checkpoint_path)
    orbax_checkpointer = ocp.PyTreeCheckpointer()
    target = training_state.normalizer_params, init_params
    (normalizer_params, init_params) = orbax_checkpointer.restore(
        restore_checkpoint_path, item=target
    )
    training_state = training_state.replace(
        normalizer_params=normalizer_params, params=init_params
    )

  training_state = jax.device_put_replicated(
      training_state,
      jax.local_devices()[:local_devices_to_use])

  if not eval_env:
    eval_env = environment
  if randomization_fn is not None:
    v_randomization_fn = functools.partial(
        randomization_fn, rng=jax.random.split(eval_key, num_eval_envs)
    )
  eval_env = wrap_for_training(
      eval_env,
      episode_length=episode_length,
      action_repeat=action_repeat,
      randomization_fn=v_randomization_fn,
  )

  evaluator = acting.Evaluator(
      eval_env,
      functools.partial(make_policy, deterministic=deterministic_eval),
      num_eval_envs=num_eval_envs,
      episode_length=episode_length,
      action_repeat=action_repeat,
      key=eval_key)

  # Run initial eval
  metrics = {}
  if process_id == 0 and num_evals > 1:
    metrics = evaluator.run_evaluation(
        _unpmap(
            (training_state.normalizer_params, training_state.params.policy)),
        training_metrics={})
    logging.info(metrics)
    progress_fn(0, metrics)

  training_metrics = {}
  training_walltime = 0
  current_step = 0
  for it in range(num_evals_after_init):
    logging.info('starting iteration %s %s', it, time.time() - xt)

    for _ in range(max(num_resets_per_eval, 1)):
      # optimization
      epoch_key, local_key = jax.random.split(local_key)
      epoch_keys = jax.random.split(epoch_key, local_devices_to_use)
      (training_state, env_state, training_metrics) = (
          training_epoch_with_timing(training_state, env_state, epoch_keys)
      )
      current_step = int(_unpmap(training_state.env_steps))

      key_envs = jax.vmap(
          lambda x, s: jax.random.split(x[0], s),
          in_axes=(0, None))(key_envs, key_envs.shape[1])
      # TODO: move extra reset logic to the AutoResetWrapper.
      env_state = reset_fn(key_envs) if num_resets_per_eval > 0 else env_state

    if process_id == 0:
      # Run evals.
      metrics = evaluator.run_evaluation(
          _unpmap(
              (training_state.normalizer_params, training_state.params.policy)),
          training_metrics)
      logging.info(metrics)
      progress_fn(current_step, metrics)
      params = _unpmap(
          (training_state.normalizer_params, training_state.params)
      )
      policy_params_fn(current_step, make_policy, params)

  total_steps = current_step
  assert total_steps >= num_timesteps

  # If there was no mistakes the training_state should still be identical on all
  # devices.
  pmap.assert_is_replicated(training_state)
  params = _unpmap(
      (training_state.normalizer_params, training_state.params.policy))
  logging.info('total steps: %s', total_steps)
  pmap.synchronize_hosts()
  return (make_policy, params, metrics)
